from ctypes import (
    c_ubyte,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import assert_cast
from .enum import (
    DX_HASH,
    EXT2_FLAGS,
    EXT4_CHKSUM,
    EXT4_DEFM,
    EXT4_ERRORS,
    EXT4_FEATURE_COMPAT,
    EXT4_FEATURE_INCOMPAT,
    EXT4_FEATURE_RO_COMPAT,
    EXT4_FS,
    EXT4_MOUNT,
    EXT4_OS,
    EXT4_REV,
    FS_ENCRYPTION_MODE,
)
from .struct import (
    Ext4Struct,
    crc32c,
)

if TYPE_CHECKING:
    from .volume import Volume


@final
class Superblock(Ext4Struct):
    __slots__ = ()

    _pack_ = 1
    # _anonymous_ = (
    #     "s_reserved_pad",
    #     "s_reserved",
    # )
    _fields_ = [
        ("s_inodes_count", c_uint32),
        ("s_blocks_count_lo", c_uint32),
        ("s_r_blocks_count_lo", c_uint32),
        ("s_free_blocks_count_lo", c_uint32),
        ("s_free_inodes_count", c_uint32),
        ("s_first_data_block", c_uint32),
        ("s_log_block_size", c_uint32),
        ("s_log_cluster_size", c_uint32),
        ("s_blocks_per_group", c_uint32),
        ("s_clusters_per_group", c_uint32),
        ("s_inodes_per_group", c_uint32),
        ("s_mtime", c_uint32),
        ("s_wtime", c_uint32),
        ("s_mnt_count", c_uint16),
        ("s_max_mnt_count", c_uint16),
        ("s_magic", c_uint16),  # 0xEF53
        ("s_state", EXT4_FS.basetype),
        ("s_errors", EXT4_ERRORS.basetype),
        ("s_minor_rev_level", c_uint16),
        ("s_lastcheck", c_uint32),
        ("s_checkinterval", c_uint32),
        ("s_creator_os", EXT4_OS.basetype),
        ("s_rev_level", EXT4_REV.basetype),
        ("s_def_resuid", c_uint16),
        ("s_def_resgid", c_uint16),
        ("s_first_ino", c_uint32),
        ("s_inode_size", c_uint16),
        ("s_block_group_nr", c_uint16),
        ("s_feature_compat", EXT4_FEATURE_COMPAT.basetype),
        ("s_feature_incompat", EXT4_FEATURE_INCOMPAT.basetype),
        ("s_feature_ro_compat", EXT4_FEATURE_RO_COMPAT.basetype),
        ("s_uuid", c_uint8 * 16),
        ("s_volume_name", c_ubyte * 16),
        ("s_last_mounted", c_ubyte * 64),
        ("s_algorithm_usage_bitmap", c_uint32),
        ("s_prealloc_blocks", c_uint8),
        ("s_prealloc_dir_blocks", c_uint8),
        ("s_reserved_gdt_blocks", c_uint16),
        ("s_journal_uuid", c_uint8 * 16),
        ("s_journal_inum", c_uint32),
        ("s_journal_dev", c_uint32),
        ("s_last_orphan", c_uint32),
        ("s_hash_seed", c_uint32 * 4),
        ("s_def_hash_version", DX_HASH.basetype),
        ("s_jnl_backup_type", c_uint8),
        ("s_desc_size", c_uint16),
        ("s_default_mount_opts", EXT4_DEFM.basetype),
        ("s_first_meta_bg", c_uint32),
        ("s_mkfs_time", c_uint32),
        ("s_jnl_blocks", c_uint32 * 17),
        ("s_blocks_count_hi", c_uint32),
        ("s_r_blocks_count_hi", c_uint32),
        ("s_free_blocks_count_hi", c_uint32),
        ("s_min_extra_isize", c_uint16),
        ("s_want_extra_isize", c_uint16),
        ("s_flags", EXT2_FLAGS.basetype),
        ("s_raid_stride", c_uint16),
        ("s_mmp_interval", c_uint16),
        ("s_mmp_block", c_uint64),
        ("s_raid_stripe_width", c_uint32),
        ("s_log_groups_per_flex", c_uint8),
        ("s_checksum_type", EXT4_CHKSUM.basetype),
        ("s_reserved_pad", c_uint16),
        ("s_kbytes_written", c_uint64),
        ("s_snapshot_inum", c_uint32),
        ("s_snapshot_id", c_uint32),
        ("s_snapshot_r_blocks_count", c_uint64),
        ("s_snapshot_list", c_uint32),
        ("s_error_count", c_uint32),
        ("s_first_error_time", c_uint32),
        ("s_first_error_ino", c_uint32),
        ("s_first_error_block", c_uint64),
        ("s_first_error_func", c_uint8 * 32),
        ("s_first_error_line", c_uint32),
        ("s_last_error_time", c_uint32),
        ("s_last_error_ino", c_uint32),
        ("s_last_error_line", c_uint32),
        ("s_last_error_block", c_uint64),
        ("s_last_error_func", c_uint8 * 32),
        ("s_mount_opts", EXT4_MOUNT.basetype * 64),
        ("s_usr_quota_inum", c_uint32),
        ("s_grp_quota_inum", c_uint32),
        ("s_overhead_blocks", c_uint32),
        ("s_backup_bgs", c_uint32 * 2),
        ("s_encrypt_algos", FS_ENCRYPTION_MODE.basetype * 4),
        ("s_encrypt_pw_salt", c_uint8 * 16),
        ("s_lpf_ino", c_uint32),
        ("s_prj_quota_inum", c_uint32),
        ("s_checksum_seed", c_uint32),
        ("s_reserved", c_uint32 * 98),
        ("s_checksum", c_uint32),
    ]

    def __init__(self, volume: "Volume", _=None) -> None:
        super().__init__(volume, 0x400)

    @property
    def has_hi(self) -> bool:
        return (self.feature_incompat & EXT4_FEATURE_INCOMPAT.IS64BIT) != 0

    @property
    def s_blocks_count(self) -> int:
        s_blocks_per_group = assert_cast(self.s_blocks_per_group, int)  # pyright: ignore[reportAny]
        s_reserved_gdt_blocks = assert_cast(self.s_reserved_gdt_blocks, int)  # pyright: ignore[reportAny]
        s_overhead_blocks = assert_cast(self.s_overhead_blocks, int)  # pyright: ignore[reportAny]
        return (
            (s_blocks_per_group) * len(self.volume.group_descriptors)
            - s_reserved_gdt_blocks
            - s_overhead_blocks
        )
        # s_blocks_count_lo = assert_cast(self.s_blocks_count_lo, int)
        # s_blocks_count_hi = assert_cast(self.s_blocks_count_hi, int)
        # if self.has_hi:
        #     return s_blocks_count_hi << 32 | s_blocks_count_lo

        # return s_blocks_count_lo

    @property
    def s_r_blocks_count(self) -> int:
        s_r_blocks_count_lo = assert_cast(self.s_r_blocks_count_lo, int)  # pyright: ignore[reportAny]
        s_r_blocks_count_hi = assert_cast(self.s_r_blocks_count_hi, int)  # pyright: ignore[reportAny]
        if self.has_hi:
            return s_r_blocks_count_hi << 32 | s_r_blocks_count_lo

        return s_r_blocks_count_lo

    @property
    def s_free_blocks_count(self) -> int:
        s_free_blocks_count_lo = assert_cast(self.s_free_blocks_count_lo, int)  # pyright: ignore[reportAny]
        s_free_blocks_count_hi = assert_cast(self.s_free_blocks_count_hi, int)  # pyright: ignore[reportAny]
        if self.has_hi:
            return s_free_blocks_count_hi << 32 | s_free_blocks_count_lo

        return s_free_blocks_count_lo

    @property
    def metadata_csum(self) -> bool:
        return self.feature_ro_compat & EXT4_FEATURE_RO_COMPAT.METADATA_CSUM != 0

    @Ext4Struct.expected_magic.getter
    def expected_magic(self) -> int:
        return 0xEF53

    @Ext4Struct.magic.getter
    def magic(self) -> int:
        s_magic = assert_cast(self.s_magic, int)  # pyright: ignore[reportAny]
        return s_magic

    @Ext4Struct.expected_checksum.getter
    def expected_checksum(self) -> int | None:
        s_checksum = assert_cast(self.s_checksum, int)  # pyright: ignore[reportAny]
        return s_checksum if self.metadata_csum else None

    @Ext4Struct.checksum.getter
    def checksum(self) -> int | None:
        return (
            crc32c(bytes(self)[: Superblock.s_checksum.offset])
            if self.metadata_csum
            else None
        )

    @property
    def feature_incompat(self) -> EXT4_FEATURE_INCOMPAT:
        return EXT4_FEATURE_INCOMPAT(self.s_feature_incompat)  # pyright: ignore[reportAny]

    @property
    def feature_compat(self) -> EXT4_FEATURE_COMPAT:
        return EXT4_FEATURE_COMPAT(self.s_feature_compat)  # pyright: ignore[reportAny]

    @property
    def feature_ro_compat(self) -> EXT4_FEATURE_RO_COMPAT:
        return EXT4_FEATURE_RO_COMPAT(self.s_feature_ro_compat)  # pyright: ignore[reportAny]

    @property
    def seed(self) -> int:
        if self.feature_incompat & EXT4_FEATURE_INCOMPAT.CSUM_SEED != 0:
            s_checksum_seed = assert_cast(self.s_checksum_seed, int)  # pyright: ignore[reportAny]
            return s_checksum_seed

        s_uuid = assert_cast(bytes(self.s_uuid), bytes)  # pyright: ignore[reportAny]
        return crc32c(s_uuid)

    @property
    def desc_size(self) -> int:
        if self.feature_incompat & EXT4_FEATURE_INCOMPAT.IS64BIT != 0:
            s_desc_size = assert_cast(self.s_desc_size, int)  # pyright: ignore[reportAny]
            return s_desc_size

        return 32



import ctypes
import errno
import warnings
from collections.abc import Callable
from ctypes import (
    LittleEndianStructure,
    addressof,
    memmove,
    sizeof,
)
from typing import (
    TYPE_CHECKING,
    cast,
)

from crcmod import (
    mkCrcFun,  # pyright: ignore[reportUnknownVariableType]
)

if TYPE_CHECKING:
    from .volume import Volume

SimpleCData = (
    ctypes.py_object
    | ctypes.c_short
    | ctypes.c_ushort
    | ctypes.c_long
    | ctypes.c_ulong
    | ctypes.c_int
    | ctypes.c_uint
    | ctypes.c_float
    | ctypes.c_double
    | ctypes.c_longdouble
    | ctypes.c_longlong
    | ctypes.c_ulonglong
    | ctypes.c_ubyte
    | ctypes.c_byte
    | ctypes.c_char
    | ctypes.c_char_p
    | ctypes.c_void_p
    | ctypes.c_bool
    | ctypes.c_wchar_p
    | ctypes.c_wchar
    | ctypes.c_uint
    | ctypes.c_uint8
    | ctypes.c_uint16
    | ctypes.c_uint32
    | ctypes.c_uint64
    | ctypes.c_int
    | ctypes.c_int8
    | ctypes.c_int16
    | ctypes.c_int32
    | ctypes.c_int64
)
crc32c = cast(Callable[..., int], mkCrcFun(0x11EDC6F41))


class MagicError(Exception):
    pass


class ChecksumError(Exception):
    pass


def to_hex(data: int | list[int] | bytes | None) -> str:
    if data is None:
        return "None"

    if isinstance(data, int):
        return f"0x{data:02X}"

    return "0x" + "".join([f"{x:02X}" for x in data])


class Ext4Struct(LittleEndianStructure):
    __slots__: tuple[str, ...] = (
        "offset",
        "volume",
    )

    def __init__(self, volume: "Volume", offset: int) -> None:
        super().__init__()
        self.volume: Volume = volume
        self.offset: int = offset
        self.read_from_volume()
        self.verify()

    @classmethod
    def field_type(cls, name: str) -> SimpleCData | None:
        assert isinstance(cls._fields_, list)
        assert not [x for x in cls._fields_ if not isinstance(x, tuple)]
        for _name, _type in cast(list[tuple[str, SimpleCData]], cls._fields_):
            if _name == name:
                return _type

        return None

    def read_from_volume(self) -> None:
        _ = self.volume.seek(self.offset)
        data = self.volume.read(sizeof(self))
        if len(data) != sizeof(self):
            raise OSError(
                errno.EIO,
                f"Short read for {type(self).__name__} at offset {self.offset}",
            )

        _ = memmove(addressof(self), data, sizeof(self))

    @property
    def size(self) -> int:
        return sizeof(self)

    @property
    def magic(self) -> int | None:
        return None

    @property
    def expected_magic(self) -> int | None:
        return None

    @property
    def checksum(self) -> int | None:
        return None

    @property
    def expected_checksum(self) -> int | None:
        return None

    @property
    def ignore_magic(self) -> bool:
        return self.volume.ignore_magic

    def verify(self) -> None:
        """
        Verify magic numbers
        """
        if self.magic == self.expected_magic:
            return

        message = (
            f"{self} magic bytes do not match! "
            f"expected={to_hex(self.expected_magic)}, "
            f"actual={to_hex(self.magic)}"
        )
        if not self.ignore_magic:
            raise MagicError(message)

        warnings.warn(message, RuntimeWarning)

    def validate(self) -> None:
        """
        Validate data checksums
        """
        if self.checksum == self.expected_checksum:
            return

        message = (
            f"{self} checksum does not match! "
            f"expected={self.expected_checksum}, "
            f"actual={self.checksum}"
        )
        if not self.volume.ignore_checksum:
            raise ChecksumError(message)

        warnings.warn(message, RuntimeWarning)



from ctypes import (
    c_uint16,
    c_uint32,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import assert_cast
from .enum import EXT4_BG
from .struct import (
    Ext4Struct,
    crc32c,
)
from .superblock import Superblock

if TYPE_CHECKING:
    from .volume import Volume


@final
class BlockDescriptor(Ext4Struct):
    __slots__ = ("bg_no",)

    _pack_ = 1
    # _anonymous_ = ("bg_reserved",)
    _fields_ = [
        ("bg_block_bitmap_lo", c_uint32),
        ("bg_inode_bitmap_lo", c_uint32),
        ("bg_inode_table_lo", c_uint32),
        ("bg_free_blocks_count_lo", c_uint16),
        ("bg_free_inodes_count_lo", c_uint16),
        ("bg_used_dirs_count_lo", c_uint16),
        ("bg_flags", EXT4_BG.basetype),
        ("bg_exclude_bitmap_lo", c_uint32),
        ("bg_block_bitmap_csum_lo", c_uint16),
        ("bg_inode_bitmap_csum_lo", c_uint16),
        ("bg_itable_unused_lo", c_uint16),
        ("bg_checksum", c_uint16),
        ("bg_block_bitmap_hi", c_uint32),
        ("bg_inode_bitmap_hi", c_uint32),
        ("bg_inode_table_hi", c_uint32),
        ("bg_free_blocks_count_hi", c_uint16),
        ("bg_free_inodes_count_hi", c_uint16),
        ("bg_used_dirs_count_hi", c_uint16),
        ("bg_itable_unused_hi", c_uint16),
        ("bg_exclude_bitmap_hi", c_uint32),
        ("bg_block_bitmap_csum_hi", c_uint16),
        ("bg_inode_bitmap_csum_hi", c_uint16),
        ("bg_reserved", c_uint32),
    ]

    def __init__(self, volume: "Volume", offset: int, bg_no: int) -> None:
        super().__init__(volume, offset)
        self.bg_no: int = bg_no

    @property
    def bg_block_bitmap(self) -> int:
        bg_block_bitmap_lo = assert_cast(self.bg_block_bitmap_lo, int)  # pyright: ignore[reportAny]
        bg_block_bitmap_hi = assert_cast(self.bg_block_bitmap_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_block_bitmap_hi << 32 | bg_block_bitmap_lo

        return bg_block_bitmap_lo

    @property
    def bg_inode_bitmap(self) -> int:
        bg_inode_bitmap_lo = assert_cast(self.bg_inode_bitmap_lo, int)  # pyright: ignore[reportAny]
        bg_inode_bitmap_hi = assert_cast(self.bg_inode_bitmap_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_inode_bitmap_hi << 32 | bg_inode_bitmap_lo

        return bg_inode_bitmap_lo

    @property
    def bg_free_blocks_count(self) -> int:
        bg_free_blocks_count_lo = assert_cast(self.bg_free_blocks_count_lo, int)  # pyright: ignore[reportAny]
        bg_free_blocks_count_hi = assert_cast(self.bg_free_blocks_count_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_free_blocks_count_hi << 16 | bg_free_blocks_count_lo

        return bg_free_blocks_count_lo

    @property
    def bg_free_inodes_count(self) -> int:
        bg_free_inodes_count_lo = assert_cast(self.bg_free_inodes_count_lo, int)  # pyright: ignore[reportAny]
        bg_free_inodes_count_hi = assert_cast(self.bg_free_inodes_count_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_free_inodes_count_hi << 16 | bg_free_inodes_count_lo

        return bg_free_inodes_count_lo

    @property
    def bg_exclude_bitmap(self) -> int:
        bg_exclude_bitmap_lo = assert_cast(self.bg_exclude_bitmap_lo, int)  # pyright: ignore[reportAny]
        bg_exclude_bitmap_hi = assert_cast(self.bg_exclude_bitmap_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_exclude_bitmap_hi << 32 | bg_exclude_bitmap_lo

        return bg_exclude_bitmap_lo

    @property
    def bg_used_dirs_count(self) -> int:
        bg_used_dirs_count_lo = assert_cast(self.bg_used_dirs_count_lo, int)  # pyright: ignore[reportAny]
        bg_used_dirs_count_hi = assert_cast(self.bg_used_dirs_count_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_used_dirs_count_hi << 16 | bg_used_dirs_count_lo

        return bg_used_dirs_count_lo

    @property
    def bg_block_bitmap_csum(self) -> int:
        bg_block_bitmap_csum_lo = assert_cast(self.bg_block_bitmap_csum_lo, int)  # pyright: ignore[reportAny]
        bg_block_bitmap_csum_hi = assert_cast(self.bg_block_bitmap_csum_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_block_bitmap_csum_hi << 16 | bg_block_bitmap_csum_lo

        return bg_block_bitmap_csum_lo

    @property
    def bg_inode_bitmap_csum(self) -> int:
        bg_inode_bitmap_csum_lo = assert_cast(self.bg_inode_bitmap_csum_lo, int)  # pyright: ignore[reportAny]
        bg_inode_bitmap_csum_hi = assert_cast(self.bg_inode_bitmap_csum_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_inode_bitmap_csum_hi << 16 | bg_inode_bitmap_csum_lo

        return bg_inode_bitmap_csum_lo

    @property
    def bg_itable_unused(self) -> int:
        bg_itable_unused_lo = assert_cast(self.bg_itable_unused_lo, int)  # pyright: ignore[reportAny]
        bg_itable_unused_hi = assert_cast(self.bg_itable_unused_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return bg_itable_unused_hi << 16 | bg_itable_unused_lo

        return bg_itable_unused_lo

    @property
    def bg_inode_table(self) -> int:
        bg_inode_table_lo = assert_cast(self.bg_inode_table_lo, int)  # pyright: ignore[reportAny]
        bg_inode_table_hi = assert_cast(self.bg_inode_table_hi, int)  # pyright: ignore[reportAny]
        if self.volume.has_hi:
            return (bg_inode_table_hi << 32) + bg_inode_table_lo

        return bg_inode_table_lo

    @property
    def superblock(self) -> Superblock:
        return self.volume.superblock

    @Ext4Struct.checksum.getter
    def checksum(self) -> int:
        csum = crc32c(self.bg_no.to_bytes(4, "little"), self.volume.seed)
        csum = crc32c(bytes(self)[: BlockDescriptor.bg_checksum.offset], csum)
        if self.volume.has_hi:
            csum = crc32c(b"\x00\x00", csum)
            csum = crc32c(
                bytes(self)[BlockDescriptor.bg_block_bitmap_hi.offset :], csum
            )
        return csum & 0xFFFF

    @Ext4Struct.expected_checksum.getter
    def expected_checksum(self) -> int:
        bg_checksum = assert_cast(self.bg_checksum, int)  # pyright: ignore[reportAny]
        return bg_checksum



from ctypes import (
    addressof,
    c_char,
    c_uint8,
    c_uint16,
    c_uint32,
    memmove,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import (
    assert_cast,
    override,
)
from .enum import EXT4_FT
from .struct import Ext4Struct

if TYPE_CHECKING:
    from .inode import Directory

EXT4_NAME_LEN = 255
EXT4_DIR_PAD = 4
EXT4_DIR_ROUND = EXT4_DIR_PAD - 1
EXT4_MAX_REC_LEN = (1 << 16) - 1


class DirectoryEntryStruct(Ext4Struct):
    __slots__: tuple[str, ...] = ("directory",)

    def __init__(self, directory: "Directory", offset: int) -> None:
        self.directory: Directory = directory
        super().__init__(directory.volume, offset)

    @override
    def read_from_volume(self) -> None:
        reader = self.directory._open()  # pyright: ignore[reportPrivateUsage]
        _ = reader.seek(self.offset)
        data = reader.read(self.size)
        # Do not guard against short reads, this happens for the last entry
        _ = memmove(addressof(self), data, len(data))


class DirectoryEntryBase(DirectoryEntryStruct):
    __slots__: tuple[str, ...] = ()

    @property
    def name_bytes(self) -> bytes:
        return bytes(self.name)[: self.name_len]  # pyright: ignore[reportAny]

    @property
    def name_str(self) -> str:
        return self.name_bytes.decode("utf-8")

    @property
    def is_fake_entry(self) -> bool:
        name_len = assert_cast(self.name_len, int)  # pyright: ignore[reportAny]
        return 0 < name_len <= 2 and self.name_bytes in (b".", b"..")


@final
class DirectoryEntry(DirectoryEntryBase):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("l_i_reserved",)
    _fields_ = [
        ("inode", c_uint32),
        ("rec_len", c_uint16),
        ("name_len", c_uint16),
        ("name", c_char * EXT4_NAME_LEN),
    ]


@final
class DirectoryEntry2(DirectoryEntryBase):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("l_i_reserved",)
    _fields_ = [
        ("inode", c_uint32),
        ("rec_len", c_uint16),
        ("name_len", c_uint8),
        ("file_type", EXT4_FT.basetype),
        ("name", c_char * EXT4_NAME_LEN),
    ]

    @DirectoryEntryBase.is_fake_entry.getter
    def is_fake_entry(self) -> bool:
        file_type = EXT4_FT(self.file_type)  # pyright: ignore[reportAny]
        return super().is_fake_entry or file_type == EXT4_FT.DIR_CSUM


@final
class DirectoryEntryTail(DirectoryEntryStruct):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("det_reserved_zero1", "det_reserved_zero2",)
    _fields_ = [
        ("det_reserved_zero1", c_uint32),
        ("det_rec_len", c_uint16),
        ("det_reserved_zero2", c_uint8),
        ("det_reserved_ft", c_uint8),  # EXT4_FT.DIR_CSUM
        ("det_checksum", c_uint32),
    ]

    @Ext4Struct.magic.getter
    def magic(self) -> int:
        det_reserved_ft = assert_cast(self.det_reserved_ft, int)  # pyright: ignore[reportAny]
        return det_reserved_ft

    @Ext4Struct.expected_magic.getter
    def expected_magic(self) -> int:
        return int(EXT4_FT.DIR_CSUM)


@final
class DirectoryEntryHash(DirectoryEntryStruct):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("det_reserved_zero1", "det_reserved_zero2",)
    _fields_ = [
        ("hash", c_uint32),
        ("minor_hash", c_uint32),
    ]



import errno
import io
import os
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import override

if TYPE_CHECKING:
    from .extent import Extent
    from .inode import Inode
    from .volume import Volume


@final
class BlockIOBlocks:
    __slots__ = (
        "_null_block",
        "blockio",
    )

    def __init__(self, blockio: "BlockIO") -> None:
        self.blockio: BlockIO = blockio
        self._null_block: bytearray = bytearray(self.block_size)

    @property
    def block_size(self) -> int:
        return self.blockio.block_size

    @property
    def volume(self) -> "Volume":
        return self.blockio.inode.volume

    def __contains__(self, ee_block: int) -> bool:
        for extent in self.blockio.extents:
            if ee_block in extent.blocks:
                return True

        return False

    def __getitem__(self, ee_block: int) -> bytearray | bytes:
        for extent in self.blockio.extents:
            if ee_block not in extent.blocks:
                continue

            return extent.blocks[ee_block]

        return self._null_block


@final
class BlockIO(io.RawIOBase):
    __slots__ = (
        "blocks",
        "cursor",
        "inode",
    )

    def __init__(self, inode: "Inode") -> None:
        super().__init__()
        self.inode: Inode = inode
        self.cursor: int = 0
        self.blocks: BlockIOBlocks = BlockIOBlocks(self)

    def __len__(self) -> int:
        return self.inode.i_size

    @property
    def extents(self) -> "list[Extent]":
        return self.inode.extents

    @property
    def block_size(self) -> int:
        return self.inode.volume.block_size

    @override
    def readable(self) -> bool:
        return True

    @override
    def seekable(self) -> bool:
        return True

    @override
    def seek(self, offset: int, mode: int = io.SEEK_SET) -> int:
        if mode == io.SEEK_CUR:
            offset += self.cursor

        elif mode == io.SEEK_END:
            offset += len(self)

        elif mode != io.SEEK_SET:
            raise NotImplementedError()

        if offset < 0:
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))

        self.cursor = offset
        return offset

    @override
    def tell(self) -> int:
        return self.cursor

    @override
    def read(self, size: int | None = -1) -> bytes:
        if size is None or size < 0:
            size = len(self) - self.cursor

        data = self.peek(size)
        self.cursor += len(data)
        if size < len(data):
            raise OSError(errno.EIO, "Unexpected EOF")

        return data

    def peek(self, size: int = 0) -> bytes:
        if self.cursor >= len(self):
            return b""

        if self.cursor + size >= len(self):
            size = len(self) - self.cursor

        start_index = self.cursor // self.block_size
        end_index = (self.cursor + size - 1) // self.block_size
        start_offset = self.cursor % self.block_size
        end_offset = ((self.cursor + size - 1) % self.block_size) + 1
        blocks_list: list[memoryview] = []

        for i in range(start_index, end_index + 1):
            block: bytes | bytearray = self.blocks[i]
            view = memoryview(block)
            if i == start_index:
                view = view[start_offset:]

            if i == end_index:
                trim = end_offset - (start_offset if i == start_index else 0)
                view = view[:trim]

            blocks_list.append(view)

        return b"".join(blocks_list)



# pyright: reportUnnecessaryTypeIgnoreComment=false
# pyright: reportIgnoreCommentWithoutRule=false
# pyright: reportUnreachable=false
# pyright: reportExplicitAny=false
# pyright: reportAny=false
import os
import sys
from typing import (
    Any,
    Protocol,
    TypeVar,
    runtime_checkable,
)

if sys.version_info < (3, 12):
    from typing_extensions import override

else:
    from typing import override


@runtime_checkable
class ReadableStream(Protocol):
    def read(self, size: int | None = -1, /) -> bytes: ...

    def tell(self) -> int: ...

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int: ...


@runtime_checkable
class PeekableStream(ReadableStream, Protocol):
    def peek(self, size: int = 0, /) -> bytes: ...


T = TypeVar("T")


def assert_cast(obj: Any, t: type[T], /) -> T:
    assert isinstance(obj, t), f"Object is: {type(obj)} not {t}"
    return obj


__all__ = ["override", "ReadableStream", "PeekableStream", "assert_cast"]



from __future__ import annotations

import errno
import io
import os
from pathlib import PurePosixPath
from uuid import UUID

from cachetools import (
    LRUCache,
    cachedmethod,
)

from ._compat import (
    PeekableStream,
    assert_cast,
)
from .blockdescriptor import BlockDescriptor
from .enum import EXT4_INO
from .inode import (
    Directory,
    Inode,
)
from .superblock import Superblock


class InvalidStreamException(Exception):
    pass


class Inodes:
    __slots__: tuple[str, ...] = (
        "_getitem_cache",
        "_group_cache",
        "_offset_cache",
        "volume",
        "__dict__",
    )

    def __init__(self, volume: Volume) -> None:
        self.volume: Volume = volume
        self._group_cache: dict[int, tuple[int, int]] = {}
        self._offset_cache: LRUCache[int, int] = LRUCache(maxsize=32)
        self._getitem_cache: LRUCache[int, Inode] = LRUCache(maxsize=32)

    @property
    def superblock(self) -> Superblock:
        return self.volume.superblock

    @property
    def block_size(self) -> int:
        return self.volume.block_size

    @cachedmethod(lambda self: self._group_cache)  # pyright: ignore[reportAny]
    def group(self, index: int) -> tuple[int, int]:
        s_inodes_per_group = assert_cast(self.superblock.s_inodes_per_group, int)  # pyright: ignore[reportAny]
        group_index = (index - 1) // s_inodes_per_group
        table_entry_index = (index - 1) % s_inodes_per_group
        return group_index, table_entry_index

    @cachedmethod(lambda self: self._offset_cache)  # pyright: ignore[reportAny]
    def offset(self, index: int) -> int:
        group_index, table_entry_index = self.group(index)
        table_offset = (
            self.volume.group_descriptors[group_index].bg_inode_table * self.block_size
        )
        s_inode_size = assert_cast(self.superblock.s_inode_size, int)  # pyright: ignore[reportAny]
        return table_offset + table_entry_index * s_inode_size

    @cachedmethod(lambda self: self._getitem_cache)  # pyright: ignore[reportAny]
    def __getitem__(self, index: int) -> Inode:
        offset = self.offset(index)
        return Inode(self.volume, offset, index)


class Volume:
    __slots__: tuple[str, ...] = (
        "cursor",
        "group_descriptors",
        "ignore_attr_name_index",
        "ignore_checksum",
        "ignore_flags",
        "ignore_magic",
        "inodes",
        "offset",
        "stream",
        "superblock",
    )

    def __init__(
        self,
        stream: PeekableStream,
        offset: int = 0,
        ignore_flags: bool = False,
        ignore_magic: bool = False,
        ignore_checksum: bool = False,
        ignore_attr_name_index: bool = False,
    ) -> None:
        errors: list[str] = []
        for name in ("read", "peek", "tell", "seek"):
            if not hasattr(stream, name):
                errors.append(f"{name} method missing")

            elif not callable(getattr(stream, name)):  # pyright: ignore[reportAny]
                errors.append(f"{name} is not a method")

        if errors:
            raise InvalidStreamException(", ".join(errors))

        self.stream: PeekableStream = stream
        self.offset: int = offset
        self.cursor: int = 0
        self.ignore_flags: bool = ignore_flags
        self.ignore_magic: bool = ignore_magic
        self.ignore_checksum: bool = ignore_checksum
        self.ignore_attr_name_index: bool = ignore_attr_name_index
        self.superblock: Superblock = Superblock(self)
        self.superblock.verify()
        self.group_descriptors: list[BlockDescriptor] = []
        block_size = self.block_size
        table_offset = (self.superblock.offset // block_size + 1) * block_size
        s_inodes_count = assert_cast(self.superblock.s_inodes_count, int)  # pyright: ignore[reportAny]
        s_inodes_per_group = assert_cast(self.superblock.s_inodes_per_group, int)  # pyright: ignore[reportAny]
        for index in range(
            0,
            (s_inodes_count + s_inodes_per_group - 1) // s_inodes_per_group,
        ):
            descriptor = BlockDescriptor(
                self,
                table_offset + (index * self.superblock.desc_size),
                index,
            )
            descriptor.verify()
            self.group_descriptors.insert(index, descriptor)

        self.inodes: Inodes = Inodes(self)

    def __len__(self) -> int:
        _ = self.stream.seek(0, io.SEEK_END)
        return self.stream.tell() - self.offset

    @property
    def bad_blocks(self) -> Inode:
        return self.inodes[EXT4_INO.BAD]

    @property
    def root(self) -> Directory:
        return assert_cast(self.inodes[EXT4_INO.ROOT], Directory)

    @property
    def user_quota(self) -> Inode:
        return self.inodes[EXT4_INO.USR_QUOTA]

    @property
    def group_quota(self) -> Inode:
        return self.inodes[EXT4_INO.GRP_QUOTA]

    @property
    def boot_loader(self) -> Inode:
        return self.inodes[EXT4_INO.BOOT_LOADER]

    @property
    def undelete_directory(self) -> Inode:
        return self.inodes[EXT4_INO.UNDEL_DIR]

    @property
    def journal(self) -> Inode:
        return self.inodes[EXT4_INO.JOURNAL]

    @property
    def has_hi(self) -> int:
        return self.superblock.has_hi

    @property
    def uuid(self) -> UUID:
        s_uuid = assert_cast(bytes(self.superblock.s_uuid), bytes)  # pyright: ignore[reportAny]
        return UUID(bytes=s_uuid)

    @property
    def seed(self) -> int:
        return self.superblock.seed

    @property
    def block_size(self) -> int:
        s_log_block_size = assert_cast(self.superblock.s_log_block_size, int)  # pyright: ignore[reportAny]
        return int(
            2
            ** (  # pyright: ignore[reportAny]
                10 + s_log_block_size
            )
        )

    def seek(self, offset: int, mode: int = io.SEEK_SET) -> int:
        if mode == io.SEEK_SET:
            seek = offset

        elif mode == io.SEEK_CUR:
            seek = self.cursor + offset

        elif mode == io.SEEK_END:
            seek = len(self) + offset

        else:
            raise NotImplementedError(f"Seek mode {mode} not implemented")

        if seek < 0 or seek > len(self):
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))

        self.cursor = seek
        return self.cursor

    def read(self, size: int) -> bytes:
        _ = self.stream.seek(self.offset + self.cursor)
        remaining = size
        data = bytearray()
        while remaining > 0:
            chunk = self.stream.read(remaining)
            if not chunk:
                break

            data.extend(chunk)
            remaining -= len(chunk)

        data = data[:size]
        self.cursor += len(data)
        return bytes(data)

    def peek(self, size: int) -> bytes:
        _ = self.stream.seek(self.offset + self.cursor)
        remaining = size
        data = bytearray()
        while remaining > 0:
            chunk = self.stream.peek(remaining)
            if not chunk:
                break

            data.extend(chunk)
            remaining -= len(chunk)
            _ = self.stream.seek(len(chunk), io.SEEK_CUR)

        data = data[:size]
        _ = self.stream.seek(self.offset + self.cursor)
        return bytes(data)

    def tell(self) -> int:
        return self.cursor

    def block_read(self, index: int, count: int = 1) -> bytes:
        assert index >= 0
        assert count > 0
        block_size = self.block_size  # Only calculate once
        _ = self.seek(index * block_size)
        return self.read(count * block_size)

    @staticmethod
    def path_tuple(path: str | bytes) -> tuple[bytes, ...]:
        if isinstance(path, bytes):
            path = path.decode("utf-8")

        return tuple(x.encode("utf-8") for x in PurePosixPath(path).parts[1:])

    def inode_at(self, path: str | bytes) -> Inode:
        return self.root.inode_at(b"/".join(self.path_tuple(path)))



import warnings
from collections.abc import Generator
from ctypes import (
    c_uint8,
    c_uint16,
    c_uint32,
    sizeof,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import (
    assert_cast,
    override,
)
from .enum import (
    EXT4_FEATURE_INCOMPAT,
    EXT4_FL,
)
from .struct import (
    Ext4Struct,
    crc32c,
)

if TYPE_CHECKING:
    from .inode import Inode


class ExtendedAttributeError(Exception):
    pass


class ExtendedAttributeBase(Ext4Struct):
    __slots__: tuple[str, ...] = (
        "data_size",
        "inode",
    )

    def __init__(self, inode: "Inode", offset: int, size: int) -> None:
        self.inode: Inode = inode
        self.data_size: int = size
        super().__init__(inode.volume, offset)


class ExtendedAttributeIBodyHeader(ExtendedAttributeBase):
    __slots__: tuple[str, ...] = ()
    _pack_ = 1  # pyright: ignore[reportUnannotatedClassAttribute]
    # _anonymous_ = ()
    _fields_ = [  # pyright: ignore[reportUnannotatedClassAttribute]
        ("h_magic", c_uint32),  # 0xEA020000
    ]

    @ExtendedAttributeBase.ignore_magic.getter
    def ignore_magic(self) -> bool:
        return False

    @ExtendedAttributeBase.magic.getter
    def magic(self) -> int:
        h_magic = assert_cast(self.h_magic, int)  # pyright: ignore[reportAny]
        return h_magic

    @ExtendedAttributeBase.expected_magic.getter
    def expected_magic(self) -> int:
        return 0xEA020000

    def value_offset(self, entry: "ExtendedAttributeEntry") -> int:
        e_value_offs = assert_cast(entry.e_value_offs, int)  # pyright: ignore[reportAny]
        return self.offset + sizeof(self) + e_value_offs

    def __iter__(self) -> Generator[tuple[str, bytes], None, None]:
        offset = self.offset + (4 * ((sizeof(self) + 3) // 4))
        i = 0
        while i < self.data_size:
            entry = ExtendedAttributeEntry(self.inode, offset + i, self.data_size - i)
            e_name_len = assert_cast(entry.e_name_len, int)  # pyright: ignore[reportAny]
            e_name_index = assert_cast(entry.e_name_index, int)  # pyright: ignore[reportAny]
            e_value_offs = assert_cast(entry.e_value_offs, int)  # pyright: ignore[reportAny]
            if (e_name_len | e_name_index | e_value_offs | entry.value_inum) == 0:
                break

            value: bytes
            e_value_size = assert_cast(entry.e_value_size, int)  # pyright: ignore[reportAny]
            if entry.value_inum != 0:
                inode = self.volume.inodes[entry.value_inum]
                if not inode.has_flag(EXT4_FL.EA_INODE):
                    message = f"Inode {inode.i_no:d} is not marked as large extended attribute value"
                    if not self.volume.ignore_flags:
                        raise ExtendedAttributeError(message)

                    warnings.warn(message, RuntimeWarning)

                # TODO determine if e_value_size or i_size are required to limit results?
                value = inode.open().read()

            elif e_value_size != 0:
                value_offset = self.value_offset(entry)
                if value_offset + e_value_size > self.offset + self.data_size:
                    value = b""

                else:
                    _ = self.volume.seek(value_offset)
                    value = self.volume.read(e_value_size)

            else:
                value = b""

            yield entry.name_str, value
            i += 4 * ((entry.size + 3) // 4)


@final
class ExtendedAttributeHeader(ExtendedAttributeIBodyHeader):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("h_reserved)
    _fields_ = [
        ("h_refcount", c_uint32),
        ("h_blocks", c_uint32),
        ("h_hash", c_uint32),
        ("h_checksum", c_uint32),
        ("h_reserved", c_uint32 * 3),
    ]

    @override
    def verify(self) -> None:
        super().verify()
        h_blocks = assert_cast(self.h_blocks, int)  # pyright: ignore[reportAny]
        if h_blocks != 1:
            raise ExtendedAttributeError(
                f"Invalid number of xattr blocks at offset 0x{self.offset:X} of inode "
                + f"{self.inode.i_no:d}: {h_blocks:d} (expected 1)"
            )

    @override
    def value_offset(self, entry: "ExtendedAttributeEntry") -> int:
        e_value_offs = assert_cast(entry.e_value_offs, int)  # pyright: ignore[reportAny]
        return self.offset + e_value_offs

    @ExtendedAttributeIBodyHeader.expected_checksum.getter
    def expected_checksum(self) -> int | None:
        h_checksum = assert_cast(self.h_checksum, int)  # pyright: ignore[reportAny]
        if not h_checksum:
            return None

        return h_checksum

    @ExtendedAttributeIBodyHeader.checksum.getter
    def checksum(self) -> int | None:
        h_checksum = assert_cast(self.h_checksum, int)  # pyright: ignore[reportAny]
        if not h_checksum:
            return None

        csum = crc32c(
            bytes(self)[: ExtendedAttributeHeader.h_checksum.offset], self.volume.seed
        )
        csum = crc32c(b"\0" * ExtendedAttributeHeader.h_checksum.size, csum)
        return crc32c(bytes(self)[: ExtendedAttributeHeader.h_reserved.offset], csum)


@final
class ExtendedAttributeEntry(ExtendedAttributeBase):
    __slots__: tuple[str, ...] = ("e_name",)

    NAME_INDICES = [
        "",
        "user.",
        "system.posix_acl_access",
        "system.posix_acl_default",
        "trusted.",
        "",
        "security.",
        "system.",
        "system.richacl",
    ]
    _pack_ = 1
    # _anonymous_ = ("h_reserved)
    _fields_ = [
        ("e_name_len", c_uint8),
        ("e_name_index", c_uint8),
        ("e_value_offs", c_uint16),
        ("e_value_inum", c_uint32),
        ("e_value_size", c_uint32),
        ("e_hash", c_uint32),
        # ("e_name", c_char * self.e_name_len),
    ]

    def __init__(self, inode: "Inode", offset: int, size: int) -> None:
        self.e_name: bytes = b""
        super().__init__(inode, offset, size)

    @override
    def read_from_volume(self) -> None:
        super().read_from_volume()
        e_name_len = assert_cast(self.e_name_len, int)  # pyright: ignore[reportAny]
        self.e_name = self.volume.stream.read(e_name_len)

    @ExtendedAttributeBase.size.getter
    def size(self) -> int:
        e_name_len = assert_cast(self.e_name_len, int)  # pyright: ignore[reportAny]
        return sizeof(self) + e_name_len

    @property
    def name_str(self) -> str:
        e_name_index = assert_cast(self.e_name_index, int)  # pyright: ignore[reportAny]
        name_index = e_name_index
        if 0 > name_index or name_index >= len(ExtendedAttributeEntry.NAME_INDICES):
            msg = f"Unknown attribute prefix {e_name_index:d}"
            if self.volume.ignore_attr_name_index:
                warnings.warn(msg, RuntimeWarning)
                name_index = 0
            else:
                raise ExtendedAttributeError(msg)

        return ExtendedAttributeEntry.NAME_INDICES[name_index] + self.e_name.decode(
            "iso-8859-2"
        )

    @property
    def value_inum(self) -> int:
        e_value_inum = assert_cast(self.e_value_inum, int)  # pyright: ignore[reportAny]
        return (
            e_value_inum
            if (
                self.volume.superblock.feature_incompat & EXT4_FEATURE_INCOMPAT.EA_INODE
            )
            != 0
            else 0
        )



from ctypes import (
    c_uint8,
    c_uint16,
    c_uint32,
)
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
    final,
)

from ._compat import override

if TYPE_CHECKING:
    from .struct import SimpleCData


def TypedEnumerationType(_type: type["SimpleCData"]):  # noqa: ANN201
    class EnumerationType(type(_type)):  # type: ignore  # pyright: ignore[reportGeneralTypeIssues, reportUntypedBaseClass] #noqa: ANN201
        def __new__(cls, name: str, bases: tuple[type, ...], data: dict[str, Any]):  # pyright: ignore[reportExplicitAny, reportUnknownParameterType]  # noqa: ANN204
            _members_: dict[str, Any]  # pyright: ignore[reportExplicitAny]
            if "_members_" not in data:
                _members_ = {}
                for key, value in data.items():  # pyright: ignore[reportAny]
                    if not key.startswith("_"):
                        _members_[key] = value

                data["_members_"] = _members_

            else:
                _members_ = cast(dict[str, Any], data["_members_"])  # pyright: ignore[reportExplicitAny]

            data["_reverse_map_"] = {v: k for k, v in _members_.items()}  # pyright: ignore[reportAny]
            cls = type(_type).__new__(cls, name, bases, data)  # pyright: ignore[reportCallIssue, reportUnknownVariableType]  # noqa: PLW0642
            for key, value in cast(dict[str, Any], cls._members_).items():  # pyright: ignore[reportExplicitAny, reportAny]
                globals()[key] = value

            return cls  # pyright: ignore[reportUnknownVariableType]

        @override
        def __repr__(self) -> str:
            return f"<Enumeration {self.__name__}>"  # pyright: ignore[reportUnknownMemberType]

    return EnumerationType


def TypedCEnumeration(_type: type["SimpleCData"]):  # noqa: ANN201
    class CEnumeration(_type, metaclass=TypedEnumerationType(_type)):  # pyright: ignore[reportGeneralTypeIssues, reportUntypedBaseClass] # noqa: ANN201,PLW1641,PLW1641
        _members_: dict[str, Any] = {}  # pyright: ignore[reportExplicitAny]
        basetype: type["SimpleCData"] = _type

        @override
        def __repr__(self) -> str:
            value = self.value  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
            return f"<{self.__class__.__name__}.{self._reverse_map_.get(value, '(unknown)')}: {value}>"  # pyright: ignore[reportUnknownMemberType]

        @override
        def __eq__(self, other: Any) -> bool:  # pyright: ignore[reportExplicitAny, reportAny]
            if isinstance(other, int):
                return self.value == other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value == other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __ne__(self, other: Any) -> bool:  # pyright: ignore[reportExplicitAny, reportAny]
            if isinstance(other, int):
                return self.value != other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value != other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __and__(self, other: Any) -> int | bool:  # pyright: ignore[reportExplicitAny, reportAny, reportGeneralTypeIssues]
            if isinstance(other, int):
                return self.value & other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value & other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __gt__(self, other: Any) -> bool:  # pyright: ignore[reportGeneralTypeIssues, reportAny, reportExplicitAny]
            if isinstance(other, int):
                return self.value > other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value > other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __lt__(self, other: Any) -> bool:  # pyright: ignore[reportGeneralTypeIssues, reportExplicitAny, reportAny]
            if isinstance(other, int):
                return self.value < other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value < other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __ge__(self, other: Any) -> bool:  # pyright: ignore[reportGeneralTypeIssues, reportExplicitAny, reportAny]
            if isinstance(other, int):
                return self.value >= other  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]

            return isinstance(self, type(other)) and self.value >= other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        @override
        def __le__(self, other: Any) -> bool:  # pyright: ignore[reportGeneralTypeIssues, reportAny, reportExplicitAny]
            if isinstance(other, int):
                return self.value <= other  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

            return isinstance(self, type(other)) and self.value <= other.value  # pyright: ignore[reportUnknownVariableType, reportAny, reportUnknownMemberType]

        # TODO Add the rest
        #  See https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types

    return CEnumeration


@final
class EXT4_INO(TypedCEnumeration(c_uint32)):
    BAD = 1  # Bad blocks inode
    ROOT = 2  # Root inode
    USR_QUOTA = 3  # User quota inode
    GRP_QUOTA = 4  # Group quota inode
    BOOT_LOADER = 5  # Boot loader inode
    UNDEL_DIR = 6  # Undelete directory inode
    RESIZE = 7  # Reserved group descriptors inode
    JOURNAL = 8  # Journal inode
    GOOD_OLD_FIRST = 11


@final
class EXT4_FS(TypedCEnumeration(c_uint16)):
    VALID = 0x0001  # Unmounted cleanly
    ERROR = 0x0002  # Errors detected
    ORPHAN = 0x0004  # Orphans being recovered
    FC_REPLAY = 0x0020  # Fast commit replay ongoing


@final
class EXT4_ERRORS(TypedCEnumeration(c_uint16)):
    DEFAULT = CONTINUE = 1  # Continue execution
    RO = 2  # Remount fs read-only
    PANIC = 3  # Panic


@final
class EXT4_OS(TypedCEnumeration(c_uint32)):
    LINUX = 0
    HURD = 1
    MASIX = 2
    FREEBSD = 3
    LITES = 4


@final
class EXT4_REV(TypedCEnumeration(c_uint32)):
    EXT4_GOOD_OLD_REV = 0  # The good old (original) format
    EXT4_DYNAMIC_REV = 1  # V2 format w/ dynamic inode sizes
    EXT4_MAX_SUPP_REV = EXT4_DYNAMIC_REV


@final
class EXT4_FEATURE_COMPAT(TypedCEnumeration(c_uint32)):
    DIR_PREALLOC = 0x0001
    IMAGIC_INODES = 0x0002
    HAS_JOURNAL = 0x0004
    EXT_ATTR = 0x0008
    RESIZE_INODE = 0x0010
    DIR_INDEX = 0x0020
    SPARSE_SUPER2 = 0x0200
    FAST_COMMIT = 0x0400
    STABLE_INODES = 0x0800
    ORPHAN_FILE = 0x1000  # Orphan file exists


@final
class EXT4_FEATURE_INCOMPAT(TypedCEnumeration(c_uint32)):
    COMPRESSION = 0x0001
    FILETYPE = 0x0002
    RECOVER = 0x0004  # Needs recovery
    JOURNAL_DEV = 0x0008  # Journal device
    META_BG = 0x0010
    EXTENTS = 0x0040  # extents support
    IS64BIT = 0x0080
    MMP = 0x0100
    FLEX_BG = 0x0200
    EA_INODE = 0x0400  # EA in inode
    DIRDATA = 0x1000  # data in dirent
    CSUM_SEED = 0x2000
    LARGEDIR = 0x4000  # >2GB or 3-lvl htree
    INLINE_DATA = 0x8000  # data in inode
    ENCRYPT = 0x10000
    CASEFOLD = 0x20000


@final
class EXT4_FEATURE_RO_COMPAT(TypedCEnumeration(c_uint32)):
    SPARSE_SUPER = 0x0001
    LARGE_FILE = 0x0002
    BTREE_DIR = 0x0004
    HUGE_FILE = 0x0008
    GDT_CSUM = 0x0010
    DIR_NLINK = 0x0020
    EXTRA_ISIZE = 0x0040
    QUOTA = 0x0100
    BIGALLOC = 0x0200
    METADATA_CSUM = 0x0400
    READONLY = 0x1000
    PROJECT = 0x2000
    VERITY = 0x8000
    ORPHAN_PRESENT = 0x10000  # Orphan file may be non-empty


@final
class DX_HASH(TypedCEnumeration(c_uint8)):
    LEGACY = 0
    HALF_MD4 = 1
    TEA = 2
    LEGACY_UNSIGNED = 3
    HALF_MD4_UNSIGNED = 4
    TEA_UNSIGNED = 5
    SIPHASH = 6


@final
class EXT4_DEFM(TypedCEnumeration(c_uint32)):
    DEBUG = 0x0001
    BSDGROUPS = 0x0002
    XATTR_USER = 0x0004
    ACL = 0x0008
    UID16 = 0x0010
    JMODE = 0x0060
    JMODE_DATA = 0x0020
    JMODE_ORDERED = 0x0040
    JMODE_WBACK = 0x0060
    NOBARRIER = 0x0100
    BLOCK_VALIDITY = 0x0200
    DISCARD = 0x0400
    NODELALLOC = 0x0800


@final
class EXT2_FLAGS(TypedCEnumeration(c_uint32)):
    SIGNED_HASH = 0x0001  # Signed dirhash in use
    UNSIGNED_HASH = 0x0002  # Unsigned dirhash in use
    TEST_FILESYS = 0x0004  # to test development code


@final
class EXT4_CHKSUM(TypedCEnumeration(c_uint8)):
    CRC32C = 1


@final
class EXT4_MOUNT(TypedCEnumeration(c_uint8)):
    NO_MBCACHE = 0x00001  # Do not use mbcache
    GRPID = 0x00004  # Create files with directory's group
    DEBUG = 0x00008  # Some debugging messages
    ERRORS_CONT = 0x00010  # Continue on errors
    ERRORS_RO = 0x00020  # Remount fs ro on errors
    ERRORS_PANIC = 0x00040  # Panic on errors
    ERRORS_MASK = 0x00070
    MINIX_DF = 0x00080  # Mimics the Minix statfs
    NOLOAD = 0x00100  # Don't use existing journa
    DAX_ALWAYS = 0x00200  # Direct Access
    DATA_FLAGS = 0x00C00  # Mode for data writes:
    JOURNAL_DATA = 0x00400  # Write data to journal
    ORDERED_DATA = 0x00800  # Flush data before commit
    WRITEBACK_DATA = 0x00C00  # No data ordering
    UPDATE_JOURNAL = 0x01000  # Update the journal format
    NO_UID32 = 0x02000  # Disable 32-bit UIDs
    XATTR_USER = 0x04000  # Extended user attributes
    POSIX_ACL = 0x08000  # POSIX Access Control Lists
    NO_AUTO_DA_ALLOC = 0x10000  # No auto delalloc mapping
    BARRIER = 0x20000  # Use block barriers
    QUOTA = 0x40000  # Some quota option set
    USRQUOTA = 0x80000  # "old" user quota, enable enforcement for hidden quota files
    GRPQUOTA = 0x100000  # "old" group quota, enable enforcement for hidden quota files
    PRJQUOTA = 0x200000  # Enable project quota enforcement
    DIOREAD_NOLOCK = 0x400000  # Enable support for dio read nolocking
    JOURNAL_CHECKSUM = 0x800000  # Journal checksums
    JOURNAL_ASYNC_COMMIT = 0x1000000  # Journal Async Commit
    WARN_ON_ERROR = 0x2000000  # Trigger WARN_ON on error
    NO_PREFETCH_BLOCK_BITMAPS = 0x4000000
    DELALLOC = 0x8000000  # Delalloc support
    DATA_ERR_ABORT = 0x10000000  # Abort on file data write
    BLOCK_VALIDITY = 0x20000000  # Block validity checking
    DISCARD = 0x40000000  # Issue DISCARD requests
    INIT_INODE_TABLE = 0x80000000  # Initialize uninitialized itables


@final
class EXT4_MOUNT2(TypedCEnumeration(c_uint8)):
    EXPLICIT_DELALLOC = 0x00000001  # User explicitly specified delalloc
    STD_GROUP_SIZE = 0x00000002  # We have standard group size of blocksize * 8 blocks
    HURD_COMPAT = 0x00000004  # Support HURD-castrated file systems
    EXPLICIT_JOURNAL_CHECKSUM = 0x00000008  # User explicitly specified journal checksum
    JOURNAL_FAST_COMMIT = 0x00000010  # Journal fast commit
    DAX_NEVER = 0x00000020  # Do not allow Direct Access
    DAX_INODE = 0x00000040  # For printing options only
    MB_OPTIMIZE_SCAN = 0x00000080  # Optimize group scanning in mballoc
    ABORT = 0x00000100  # Abort filesystem


@final
class FS_ENCRYPTION_MODE(TypedCEnumeration(c_uint8)):
    INVALID = 0  # never used
    AES_256_XTS = 1
    AES_256_GCM = 2  # never used
    AES_256_CBC = 3  # never used
    AES_256_CTS = 4
    AES_128_CBC = 5
    AES_128_CTS = 6
    ADIANTUM = 9


@final
class EXT4_BG(TypedCEnumeration(c_uint16)):
    INODE_UNINIT = 0x0001  # Inode table/bitmap not in use
    BLOCK_UNINIT = 0x0002  # Block bitmap not in use
    INODE_ZEROED = 0x0004  # On-disk itable initialized to zero


@final
class MODE(TypedCEnumeration(c_uint16)):
    IXOTH = 0x1  # (Others may execute)
    IWOTH = 0x2  # (Others may write)
    IROTH = 0x4  # (Others may read)
    IXGRP = 0x8  # (Group members may execute)
    IWGRP = 0x10  # (Group members may write)
    IRGRP = 0x20  # (Group members may read)
    IXUSR = 0x40  # (Owner may execute)
    IWUSR = 0x80  # (Owner may write)
    IRUSR = 0x100  # (Owner may read)
    ISVTX = 0x200  # (Sticky bit)
    ISGID = 0x400  # (Set GID)
    ISUID = 0x800  # (Set UID)
    # These are mutually-exclusive file types:
    IFIFO = 0x1000  # (FIFO)
    IFCHR = 0x2000  # (Character device)
    IFDIR = 0x4000  # (Directory)
    IFBLK = 0x6000  # (Block device)
    IFREG = 0x8000  # (Regular file)
    IFLNK = 0xA000  # (Symbolic link)
    IFSOCK = 0xC000  # (Socket)


@final
class EXT4_FL(TypedCEnumeration(c_uint32)):
    SECRM = 0x00000001  # Secure deletion
    UNRM = 0x00000002  # Undelete
    COMPR = 0x00000004  # Compress file
    SYNC = 0x00000008  # Synchronous updates
    IMMUTABLE = 0x00000010  # Immutable file
    APPEND = 0x00000020  # writes to file may only append
    NODUMP = 0x00000040  # do not dump file
    NOATIME = 0x00000080  # do not update atime
    DIRTY = 0x00000100
    COMPRBLK = 0x00000200  # One or more compressed clusters
    NOCOMPR = 0x00000400  # Don't compress
    ENCRYPT = 0x00000800  # encrypted file
    INDEX = 0x00001000  # hash-indexed directory
    IMAGIC = 0x00002000  # AFS directory
    JOURNAL_DATA = 0x00004000  # file data should be journaled
    NOTAIL = 0x00008000  # file tail should not be merged
    DIRSYNC = 0x00010000  # dirsync behaviour (directories only)
    TOPDIR = 0x00020000  # Top of directory hierarchie
    HUGE_FILE = 0x00040000  # Set to each huge file
    EXTENTS = 0x00080000  # Inode uses extents
    VERITY = 0x00100000  # Verity protected inode
    EA_INODE = 0x00200000  # Inode used for large EA
    DAX = 0x02000000  # Inode is DAX
    INLINE_DATA = 0x10000000  # Inode has inline data.
    PROJINHERIT = 0x20000000  # Create with parents projid
    CASEFOLD = 0x40000000  # Casefolded directory
    RESERVED = 0x80000000  # reserved for ext4 lib


@final
class EXT4_FT(TypedCEnumeration(c_uint8)):
    UNKNOWN = 0
    REG_FILE = 1
    DIR = 2
    CHRDEV = 3
    BLKDEV = 4
    FIFO = 5
    SOCK = 6
    SYMLINK = 7
    MAX = 8
    DIR_CSUM = 0xDE



from __future__ import annotations

import errno
import io
import os
import warnings
from collections.abc import Generator
from ctypes import (
    LittleEndianStructure,
    LittleEndianUnion,
    c_uint16,
    c_uint32,
    sizeof,
)
from typing import (
    TYPE_CHECKING,
    Any,
    final,
)

from cachetools import (
    LRUCache,
    cachedmethod,
)

from ._compat import (
    ReadableStream,
    assert_cast,
    override,
)
from .block import BlockIO
from .directory import (
    EXT4_DIR_ROUND,
    DirectoryEntry,
    DirectoryEntry2,
    DirectoryEntryHash,
)
from .enum import (
    EXT4_FEATURE_INCOMPAT,
    EXT4_FL,
    EXT4_FT,
    EXT4_OS,
    MODE,
)
from .extent import (
    Extent,
    ExtentHeader,
    ExtentIndex,
    ExtentTree,
)
from .htree import (
    DXRoot,
    HtreeHashError,
)
from .struct import (
    Ext4Struct,
    MagicError,
    crc32c,
)
from .superblock import Superblock
from .xattr import (
    ExtendedAttributeHeader,
    ExtendedAttributeIBodyHeader,
)

if TYPE_CHECKING:
    from .volume import Volume


class OpenDirectoryError(Exception):
    pass


class InodeError(Exception):
    pass


class MalformedInodeError(Exception):
    pass


@final
class Linux1(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("l_i_version", c_uint32),
    ]


@final
class Hurd1(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("h_i_translator", c_uint32),
    ]


@final
class Masix1(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("m_i_reserved1",)
    _fields_ = [
        ("m_i_reserved1", c_uint32),
    ]


@final
class Osd1(LittleEndianUnion):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("linux1", Linux1),
        ("hurd1", Hurd1),
        ("masix1", Masix1),
    ]


@final
class Linux2(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("l_i_reserved",)
    _fields_ = [
        ("l_i_blocks_high", c_uint16),
        ("l_i_file_acl_high", c_uint16),
        ("l_i_uid_high", c_uint16),
        ("l_i_gid_high", c_uint16),
        ("l_i_checksum_lo", c_uint16),
        ("l_i_reserved", c_uint16),
    ]


@final
class Hurd2(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("h_i_reserved1",)
    _fields_ = [
        ("h_i_reserved1", c_uint16),
        ("h_i_mode_high", c_uint16),
        ("h_i_uid_high", c_uint16),
        ("h_i_gid_high", c_uint16),
        ("h_i_author", c_uint32),
    ]


@final
class Masix2(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("h_i_reserved1", "m_i_reserved2")
    _fields_ = [
        ("h_i_reserved1", c_uint16),
        ("m_i_file_acl_high", c_uint16),
        ("m_i_reserved2", c_uint32 * 2),
    ]


@final
class Osd2(LittleEndianUnion):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("linux2", Linux2),
        ("hurd2", Hurd2),
        ("masix2", Masix2),
    ]


class Inode(Ext4Struct):
    EXT4_GOOD_OLD_INODE_SIZE: int = 128
    EXT2_GOOD_OLD_INODE_SIZE: int = 128
    __slots__: tuple[str, ...] = (
        "i_no",
        "tree",
    )
    _pack_ = 1  # pyright: ignore[reportUnannotatedClassAttribute]
    _fields_ = [  # pyright: ignore[reportUnannotatedClassAttribute]
        ("i_mode", MODE.basetype),
        ("i_uid", c_uint16),
        ("i_size_lo", c_uint32),
        ("i_atime", c_uint32),
        ("i_ctime", c_uint32),
        ("i_mtime", c_uint32),
        ("i_dtime", c_uint32),
        ("i_gid", c_uint16),
        ("i_links_count", c_uint16),
        ("i_blocks_lo", c_uint32),
        ("i_flags", EXT4_FL.basetype),
        ("osd1", Osd1),
        ("i_block", c_uint32 * 15),
        ("i_generation", c_uint32),
        ("i_file_acl_lo", c_uint32),
        ("i_size_high", c_uint32),
        ("i_obso_faddr", c_uint32),
        ("osd2", Osd2),
        ("i_extra_isize", c_uint16),
        ("i_checksum_hi", c_uint16),
        ("i_ctime_extra", c_uint32),
        ("i_mtime_extra", c_uint32),
        ("i_atime_extra", c_uint32),
        ("i_crtime", c_uint32),
        ("i_crtime_extra", c_uint32),
        ("i_version_hi", c_uint32),
        ("i_projid", c_uint32),
    ]

    @classmethod
    def get_file_type(cls, volume: Volume, offset: int) -> EXT4_FT:
        _ = volume.seek(offset + Inode.i_mode.offset)
        file_type = MODE(
            int.from_bytes(volume.read(Inode.i_mode.size), "little") & 0xF000
        )
        match file_type:
            case MODE.IFIFO:
                return EXT4_FT.FIFO  # pyright: ignore[reportReturnType]

            case MODE.IFCHR:
                return EXT4_FT.CHRDEV  # pyright: ignore[reportReturnType]

            case MODE.IFDIR:
                return EXT4_FT.DIR  # pyright: ignore[reportReturnType]

            case MODE.IFBLK:
                return EXT4_FT.BLKDEV  # pyright: ignore[reportReturnType]

            case MODE.IFREG:
                return EXT4_FT.REG_FILE  # pyright: ignore[reportReturnType]

            case MODE.IFLNK:
                return EXT4_FT.SYMLINK  # pyright: ignore[reportReturnType]

            case MODE.IFSOCK:
                return EXT4_FT.SOCK  # pyright: ignore[reportReturnType]

            case _:
                return EXT4_FT.UNKNOWN  # pyright: ignore[reportReturnType]

    def __new__(cls, volume: Volume, offset: int, i_no: int) -> Inode:
        if cls is not Inode:
            return super().__new__(cls)

        file_type = cls.get_file_type(volume, offset)
        match file_type:
            case EXT4_FT.FIFO:
                return super().__new__(Fifo)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.DIR:
                return super().__new__(Directory)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.REG_FILE:
                return super().__new__(File)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.SYMLINK:
                return super().__new__(SymbolicLink)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.CHRDEV:
                return super().__new__(CharacterDevice)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.BLKDEV:
                return super().__new__(BlockDevice)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.SOCK:
                return super().__new__(Socket)  # pyright: ignore[reportArgumentType]

            case EXT4_FT.UNKNOWN:
                return super().__new__(UnknownInode)  # pyright: ignore[reportArgumentType]

            case _:
                raise InodeError(f"Unknown file type 0x{file_type:X}")

    def __init__(self, volume: Volume, offset: int, i_no: int) -> None:
        self.i_no: int = i_no
        super().__init__(volume, offset)
        self.tree: ExtentTree | None = ExtentTree(self)

    @property
    def extra_inode_data(self) -> bytes:
        if not self.has_hi:
            return b""

        size = sizeof(self)
        assert size == self.EXT2_GOOD_OLD_INODE_SIZE + self.i_extra_isize  # pyright: ignore[reportAny]
        _ = self.volume.seek(self.offset + size)
        return self.volume.read(self.superblock.s_inode_size - size)  # pyright: ignore[reportAny]

    @property
    def superblock(self) -> Superblock:
        return self.volume.superblock

    @property
    def block_size(self) -> int:
        return self.volume.block_size

    @property
    def i_size(self) -> int:
        i_size_lo = assert_cast(self.i_size_lo, int)  # pyright: ignore[reportAny]
        i_size_high = assert_cast(self.i_size_high, int)  # pyright: ignore[reportAny]
        return i_size_high << 32 | i_size_lo

    @property
    def i_file_acl(self) -> int:
        i_file_acl_lo = assert_cast(self.i_file_acl_lo, int)  # pyright: ignore[reportAny]
        l_i_file_acl_high = assert_cast(self.osd2.linux2.l_i_file_acl_high, int)  # pyright: ignore[reportAny]
        return l_i_file_acl_high << 32 | i_file_acl_lo

    @property
    def has_hi(self) -> bool:
        s_inode_size = assert_cast(self.superblock.s_inode_size, int)  # pyright: ignore[reportAny]
        return s_inode_size > self.EXT2_GOOD_OLD_INODE_SIZE

    @property
    def fits_in_hi(self) -> bool:
        i_extra_isize = assert_cast(self.i_extra_isize, int)  # pyright: ignore[reportAny]
        return (
            self.has_hi
            and Inode.i_checksum_hi.offset + Inode.i_checksum_hi.size
            <= self.EXT2_GOOD_OLD_INODE_SIZE + i_extra_isize
        )

    @property
    def seed(self) -> int:
        seed = crc32c(self.i_no.to_bytes(4, "little"), self.volume.seed)
        return crc32c(
            self.i_generation.to_bytes(Inode.i_generation.size, "little"),  # pyright: ignore[reportAny]
            seed,
        )

    @Ext4Struct.checksum.getter
    def checksum(self) -> int | None:
        s_creator_os: EXT4_OS = EXT4_OS(self.superblock.s_creator_os)  # pyright: ignore[reportAny]
        if s_creator_os != EXT4_OS.LINUX:
            return None

        data = bytes(self)
        checksum_offset = (
            Inode.osd2.offset + Osd2.linux2.offset + Linux2.l_i_checksum_lo.offset
        )
        checksum_size = Linux2.l_i_checksum_lo.size
        csum = crc32c(data[:checksum_offset], self.seed)
        csum = crc32c(b"\0" * checksum_size, csum)
        csum = crc32c(
            data[checksum_offset + checksum_size : self.EXT2_GOOD_OLD_INODE_SIZE],
            csum,
        )
        if self.has_hi:
            offset = Inode.i_checksum_hi.offset
            csum = crc32c(data[self.EXT2_GOOD_OLD_INODE_SIZE : offset], csum)
            if self.fits_in_hi:
                csum = crc32c(b"\0" * Inode.i_checksum_hi.size, csum)
                offset += Inode.i_checksum_hi.size

            csum = crc32c(
                data[offset:],
                csum,
            )
            s_inode_size = assert_cast(self.superblock.s_inode_size, int)  # pyright: ignore[reportAny]
            if s_inode_size - len(data) > 0:
                csum = crc32c(self.extra_inode_data, csum)

        if not self.has_hi:
            csum &= 0xFFFF

        return csum

    @Ext4Struct.expected_checksum.getter
    def expected_checksum(self) -> int | None:
        s_creator_os = EXT4_OS(self.superblock.s_creator_os)  # pyright: ignore[reportAny]
        if s_creator_os != EXT4_OS.LINUX:
            return None

        provided_csum: int = 0
        l_i_checksum_lo = assert_cast(self.osd2.linux2.l_i_checksum_lo, int)  # pyright: ignore[reportAny]
        provided_csum |= l_i_checksum_lo
        if self.fits_in_hi:
            i_checksum_hi = assert_cast(self.i_checksum_hi, int)  # pyright: ignore[reportAny]
            provided_csum |= i_checksum_hi << 16

        return provided_csum

    @override
    def validate(self) -> None:
        super().validate()
        if self.tree is not None:
            self.tree.validate()

    def has_flag(self, flag: EXT4_FL | int) -> bool:
        i_flags = EXT4_FL(self.i_flags)  # pyright: ignore[reportAny]
        return (i_flags & flag) != 0

    @property
    def is_inline(self) -> bool:
        return self.has_flag(EXT4_FL.INLINE_DATA)

    @property
    def has_extents(self) -> bool:
        return self.has_flag(EXT4_FL.EXTENTS)

    @property
    def extents(self) -> list[Extent]:
        assert self.tree is not None
        return self.tree.extents

    @property
    def headers(self) -> list[ExtentHeader]:
        assert self.tree is not None
        return self.tree.headers

    @property
    def indices(self) -> list[ExtentIndex]:
        assert self.tree is not None
        return self.tree.indices

    def _open(
        self, mode: str = "rb", encoding: None = None, newline: None = None
    ) -> ReadableStream:
        if mode != "rb" or encoding is not None or newline is not None:
            raise NotImplementedError()

        if self.is_inline:
            _ = self.volume.seek(self.offset + Inode.i_block.offset)
            data = self.volume.read(self.i_size)
            return io.BytesIO(data)

        return BlockIO(self)

    def open(
        self,
        mode: str = "rb",  # pyright: ignore[reportUnusedParameter]
        encoding: None = None,  # pyright: ignore[reportUnusedParameter]
        newline: None = None,  # pyright: ignore[reportUnusedParameter]
    ) -> ReadableStream:
        raise NotImplementedError()

    @property
    def xattrs(
        self,
    ) -> Generator[tuple[str, bytes], None, None]:
        i_extra_isize = assert_cast(self.i_extra_isize, int)  # pyright: ignore[reportAny]
        inline_offset = self.offset + self.EXT2_GOOD_OLD_INODE_SIZE + i_extra_isize
        s_inode_size = assert_cast(self.superblock.s_inode_size, int)  # pyright: ignore[reportAny]
        inline_size = self.offset + s_inode_size - inline_offset
        if inline_size > sizeof(ExtendedAttributeIBodyHeader):
            try:
                header = ExtendedAttributeIBodyHeader(self, inline_offset, inline_size)
                header.verify()
                for name, value in header:
                    yield name, value

            except MagicError:
                pass

        if self.i_file_acl != 0:
            block_offset = self.i_file_acl * self.block_size
            try:
                header = ExtendedAttributeHeader(self, block_offset, self.block_size)
                header.verify()
                for name, value in header:
                    yield name, value

            except MagicError:
                if not self.superblock.ignore_magic:
                    raise


@final
class UnknownInode(Inode):
    __slots__ = ()


@final
class Fifo(Inode):
    __slots__ = ()


@final
class CharacterDevice(Inode):
    __slots__ = ()


@final
class BlockDevice(Inode):
    __slots__ = ()


@final
class Socket(Inode):
    __slots__ = ()


@final
class File(Inode):
    __slots__ = ()

    @override
    def open(
        self, mode: str = "rb", encoding: None = None, newline: None = None
    ) -> ReadableStream:
        return self._open(mode, encoding, newline)


@final
class SymbolicLink(Inode):
    __slots__ = ()

    @property
    def is_fast_symlink(self) -> bool:
        i_blocks_lo = assert_cast(self.i_blocks_lo, int)  # pyright: ignore[reportAny]
        return i_blocks_lo == 0 and not self.is_inline

    def readlink(self) -> bytes:
        if not self.is_fast_symlink:
            return self._open().read()

        if self.i_size > Inode.i_block.size:
            raise MalformedInodeError(
                f"Fast symlink target too large: {self.i_size} > {Inode.i_block.size}"
            )

        _ = self.volume.seek(self.offset + Inode.i_block.offset)
        return self.volume.read(self.i_size)


@final
class Directory(Inode):
    __slots__ = (
        "_dirents",
        "_inode_at_cache",
        "htree",
        "__dict__",
    )

    def __init__(self, volume: Volume, offset: int, i_no: int) -> None:
        super().__init__(volume, offset, i_no)
        self._inode_at_cache: LRUCache[str | bytes, Inode] = LRUCache(maxsize=32)
        self._dirents: None | list[DirectoryEntry | DirectoryEntry2] = None
        self.htree: DXRoot | None = None
        if self.is_htree:
            self.htree = DXRoot(self)

    @override
    def verify(self) -> None:
        super().verify()
        # TODO verify DirectoryEntryHash? Or should this be in validate?

    @override
    def validate(self) -> None:
        super().validate()
        # TODO validate each directory entry block with DirectoryEntryTail

    @property
    def has_filetype(self) -> bool:
        return self.superblock.feature_incompat & EXT4_FEATURE_INCOMPAT.FILETYPE != 0

    @property
    def is_htree(self) -> bool:
        return self.has_flag(EXT4_FL.INDEX)

    @property
    def is_casefolded(self) -> bool:
        return self.has_flag(EXT4_FL.CASEFOLD)

    @property
    def is_encrypted(self) -> bool:
        return self.has_flag(EXT4_FL.ENCRYPT)

    @property
    def hash_in_dirent(self) -> bool:
        return self.is_casefolded and self.is_encrypted

    def _opendir(
        self,
    ) -> Generator[DirectoryEntry | DirectoryEntry2, None, None]:
        if self._dirents is not None:
            for dirent in self._dirents:
                yield dirent

            return

        _type = DirectoryEntry2 if self.has_filetype else DirectoryEntry
        dirents: list[DirectoryEntry | DirectoryEntry2] = []
        offset: int = 0
        data = self._open().read()
        while offset < len(data):
            dirent = _type(self, offset)
            rec_len = assert_cast(dirent.rec_len, int)  # pyright: ignore[reportAny]
            if not rec_len:
                # How did this happen?
                offset += _type.name.offset  # + EXT4_DIR_ROUND
                continue

            name_len = assert_cast(dirent.name_len, int)  # pyright: ignore[reportAny]
            dirent_inode = assert_cast(dirent.inode, int)  # pyright: ignore[reportAny]
            if not dirent_inode or not name_len:
                offset += rec_len
                continue

            expected_rec_len: int = _type.name.offset + name_len + EXT4_DIR_ROUND
            if not dirent.is_fake_entry and self.hash_in_dirent:
                expected_rec_len += sizeof(DirectoryEntryHash)

            expected_rec_len &= ~EXT4_DIR_ROUND

            if rec_len < expected_rec_len:
                warnings.warn(
                    "Directory entry is too small for name length"
                    + f", expected={expected_rec_len}"
                    + f", actual={rec_len}",
                    RuntimeWarning,
                    stacklevel=2,
                )
                break

            offset += rec_len
            dirents.append(dirent)
            yield dirent

        self._dirents = dirents

    def _is_valid_file_type(self, file_type: EXT4_FT) -> bool:
        return file_type != EXT4_FT.UNKNOWN and file_type < EXT4_FT.MAX

    def _get_file_type(self, dirent: DirectoryEntry | DirectoryEntry2) -> EXT4_FT:
        dirent_inode = assert_cast(dirent.inode, int)  # pyright: ignore[reportAny]
        offset = self.volume.inodes.offset(dirent_inode)
        file_type = self.get_file_type(self.volume, offset)
        if not self._is_valid_file_type(file_type):
            raise OpenDirectoryError(
                f"Unexpected file type {file_type} for inode {dirent_inode}"
            )

        return file_type

    def opendir(
        self,
    ) -> Generator[tuple[DirectoryEntry | DirectoryEntry2, EXT4_FT], Any, None]:  # pyright: ignore[reportExplicitAny]
        for dirent in self._opendir():
            if not isinstance(dirent, DirectoryEntry2):
                file_type = self._get_file_type(dirent)
                yield dirent, file_type
                continue

            file_type = EXT4_FT(dirent.file_type)  # pyright: ignore[reportAny]
            if file_type == EXT4_FT.DIR_CSUM:
                continue

            if self._is_valid_file_type(file_type):
                yield dirent, file_type
                continue

            raise OpenDirectoryError(f"Unexpected file type: {file_type}")

    @cachedmethod(lambda self: self._inode_at_cache)  # pyright: ignore[reportAny]
    def inode_at(self, path: str | bytes) -> Inode:
        if (isinstance(path, str) and path.startswith("/")) or (
            isinstance(path, bytes) and path.startswith(b"/")
        ):
            return self.volume.inode_at(path)

        if isinstance(path, bytes):
            path = path.decode("utf-8")

        paths = list(self.volume.path_tuple(f"/{path}"))
        cwd = self
        if not paths:
            return cwd

        while paths:
            if not isinstance(cwd, Directory):
                raise OSError(errno.ENOTDIR, os.strerror(errno.ENOTDIR))

            name = paths.pop(0)
            inode = None

            if cwd.is_htree and cwd.htree is not None:
                try:
                    inode_no = cwd.htree.lookup(name)
                    if inode_no is not None:
                        inode = self.volume.inodes[inode_no]

                except HtreeHashError:
                    pass

            if inode is None:
                for dirent, _ in cwd.opendir():
                    if dirent.name_bytes == name:
                        dirent_inode = assert_cast(dirent.inode, int)  # pyright: ignore[reportAny]
                        inode = self.volume.inodes[dirent_inode]
                        break

            if inode is None:
                raise FileNotFoundError(path)

            cwd = inode

        return cwd



import errno
import warnings
from collections.abc import Iterator
from ctypes import (
    LittleEndianStructure,
    addressof,
    c_char,
    c_uint8,
    c_uint16,
    c_uint32,
    memmove,
    sizeof,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import (
    assert_cast,
    override,
)
from .block import BlockIO
from .enum import DX_HASH
from .struct import (
    Ext4Struct,
    MagicError,
)

if TYPE_CHECKING:
    from .inode import Directory
    from .volume import Volume


class HtreeHashError(Exception):
    pass


class LittleEndianStructureWithVolume(LittleEndianStructure):
    __slots__: tuple[str, ...] = ("_volume",)

    def __init__(self) -> None:
        super().__init__()
        self._volume: Volume | None = None

    @property
    def volume(self) -> "Volume":
        assert self._volume is not None
        return self._volume

    @volume.setter
    def volume(self, volume: "Volume") -> None:
        self._volume = volume


@final
class DotDirectoryEntry2(LittleEndianStructureWithVolume):
    __slots__ = ()

    _pack_ = 1
    # _anonymous_ = ()
    _fields_ = [
        ("inode", c_uint32),
        ("rec_len", c_uint16),
        ("name_len", c_uint8),
        ("file_type", c_uint8),
        ("name", c_char * 4),  # b".\0\0\0" or b"..\0\0"
    ]

    def verify(self) -> None:
        name = assert_cast(self.name, bytes)  # pyright: ignore[reportAny]
        if name in (b".", b".."):
            return

        message = f"{self} dot or dotdot entry name invalid! actual={name}"
        assert self.volume is not None
        if not self.volume.ignore_magic:
            raise MagicError(message)

        warnings.warn(
            message,
            RuntimeWarning,
            stacklevel=2,
        )


@final
class DXRootInfo(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("reserved_zero")
    _fields_ = [
        ("reserved_zero", c_uint32),
        ("hash_version", DX_HASH.basetype),
        ("info_length", c_uint8),
        ("indirect_levels", c_uint8),
        ("unused_flags", c_uint8),
    ]


class DXBase(Ext4Struct):
    __slots__: tuple[str, ...] = ("directory",)

    def __init__(self, directory: "Directory", offset: int) -> None:
        self.directory: Directory = directory
        super().__init__(directory.volume, offset)

    @override
    def read_from_volume(self) -> None:
        reader = self.directory._open()  # pyright: ignore[reportPrivateUsage]
        _ = reader.seek(self.offset)
        data = reader.read(sizeof(self))
        if len(data) != sizeof(self):
            raise OSError(
                errno.EIO,
                f"Short read for {type(self).__name__} at offset {self.offset}",
            )

        _ = memmove(addressof(self), data, sizeof(self))


@final
class DXEntry(DXBase):
    __slots__ = (
        "index",
        "parent",
    )

    _pack_ = 1
    # _anonymous_ = ("")
    _fields_ = [
        ("hash", c_uint32),
        ("block", c_uint32),
    ]

    def __init__(self, parent: "DXEntriesBase", index: int) -> None:
        self.index: int = index
        self.parent: DXEntriesBase = parent
        super().__init__(
            parent.directory,
            parent.offset + parent.size + index * parent.info_length,
        )


@final
class DXEntries:
    __slots__ = (
        "_cache",
        "base",
    )

    def __init__(self, base: "DXEntriesBase") -> None:
        self.base: DXEntriesBase = base
        self._cache: dict[int, DXEntry] = {}

    def __contains__(self, index: int) -> bool:
        return 0 <= index < self.base.count  # pyright: ignore[reportAny]

    def __len__(self) -> int:
        """Length of entries minus the DXTail"""
        return self.base.count - 1  # pyright: ignore[reportAny]

    def __iter__(self) -> Iterator[DXEntry]:
        for index in range(len(self)):
            yield self[index]

    def __getitem__(self, index: int) -> DXEntry:
        if index not in self:
            raise KeyError()

        entry = self._cache.get(index, None)
        if entry is None:
            entry = DXEntry(self.base, index)
            self._cache[index] = entry

        return entry


class DXEntriesBase(DXBase):
    __slots__: tuple[str, ...] = (
        "entries",
        "parent",
    )

    def __init__(self, directory: "Directory", offset: int) -> None:
        super().__init__(directory, offset)
        self.parent: DXRoot | None = directory.htree
        self.entries: DXEntries = DXEntries(self)

    @property
    def info_length(self) -> int:
        parent = self
        while not isinstance(parent, DXRoot):
            parent = assert_cast(parent.parent, DXEntriesBase)

        dx_root_info = assert_cast(parent.dx_root_info, DXRootInfo)  # pyright: ignore[reportAny]
        return assert_cast(dx_root_info.info_length, int)  # pyright: ignore[reportAny]


def str2hashbuf(
    name: bytes, length: int, num: int, signed_hash: bool = False
) -> list[int]:
    """Convert name bytes to array of 32-bit words (LSB-first)."""
    pad: int = length | (length << 8)
    pad |= pad << 16
    pad &= 0xFFFFFFFF

    buf: list[int] = []
    val: int = pad
    length = min(length, num * 4)
    for i in range(length):
        byte_val = sign_extended_byte(signed_hash, name[i])
        val = (byte_val + (val << 8)) & 0xFFFFFFFF
        if i % 4 == 3:
            buf.append(val)
            val = pad

    if len(buf) < num:
        buf.append(val)

    while len(buf) < num:
        buf.append(pad)

    return buf[:num]


def sign_extended_byte(signed_hash: bool, value: int) -> int:
    if signed_hash and value > 127:
        return value - 256

    return value


def half_md4_transform(buf: list[int], inp: list[int]) -> None:
    """
    Args:
        buf: Input/output [a, b, c, d] (modified in place)
        inp: 8 input words from str2hashbuf
    """

    def F(x: int, y: int, z: int) -> int:
        return (z) ^ ((x) & ((y) ^ (z)))

    def G(x: int, y: int, z: int) -> int:
        return ((x) & (y)) + (((x) ^ (y)) & (z))

    def H(x: int, y: int, z: int) -> int:
        return (x) ^ (y) ^ (z)

    def rol32(val: int, shift: int) -> int:
        """32-bit rotate left."""
        shift &= 31
        return ((val << shift) | (val >> (32 - shift))) & 0xFFFFFFFF

    a: int = buf[0]
    b: int = buf[1]
    c: int = buf[2]
    d: int = buf[3]

    k: int = 0
    a = rol32((a + F(b, c, d) + inp[0] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + F(a, b, c) + inp[1] + k) & 0xFFFFFFFF, 7)
    c = rol32((c + F(d, a, b) + inp[2] + k) & 0xFFFFFFFF, 11)
    b = rol32((b + F(c, d, a) + inp[3] + k) & 0xFFFFFFFF, 19)
    a = rol32((a + F(b, c, d) + inp[4] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + F(a, b, c) + inp[5] + k) & 0xFFFFFFFF, 7)
    c = rol32((c + F(d, a, b) + inp[6] + k) & 0xFFFFFFFF, 11)
    b = rol32((b + F(c, d, a) + inp[7] + k) & 0xFFFFFFFF, 19)

    k = 0x5A827999
    a = rol32((a + G(b, c, d) + inp[1] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + G(a, b, c) + inp[3] + k) & 0xFFFFFFFF, 5)
    c = rol32((c + G(d, a, b) + inp[5] + k) & 0xFFFFFFFF, 9)
    b = rol32((b + G(c, d, a) + inp[7] + k) & 0xFFFFFFFF, 13)
    a = rol32((a + G(b, c, d) + inp[0] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + G(a, b, c) + inp[2] + k) & 0xFFFFFFFF, 5)
    c = rol32((c + G(d, a, b) + inp[4] + k) & 0xFFFFFFFF, 9)
    b = rol32((b + G(c, d, a) + inp[6] + k) & 0xFFFFFFFF, 13)

    k = 0x6ED9EBA1
    a = rol32((a + H(b, c, d) + inp[3] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + H(a, b, c) + inp[7] + k) & 0xFFFFFFFF, 9)
    c = rol32((c + H(d, a, b) + inp[2] + k) & 0xFFFFFFFF, 11)
    b = rol32((b + H(c, d, a) + inp[6] + k) & 0xFFFFFFFF, 15)
    a = rol32((a + H(b, c, d) + inp[1] + k) & 0xFFFFFFFF, 3)
    d = rol32((d + H(a, b, c) + inp[5] + k) & 0xFFFFFFFF, 9)
    c = rol32((c + H(d, a, b) + inp[0] + k) & 0xFFFFFFFF, 11)
    b = rol32((b + H(c, d, a) + inp[4] + k) & 0xFFFFFFFF, 15)

    buf[0] = (buf[0] + a) & 0xFFFFFFFF
    buf[1] = (buf[1] + b) & 0xFFFFFFFF
    buf[2] = (buf[2] + c) & 0xFFFFFFFF
    buf[3] = (buf[3] + d) & 0xFFFFFFFF


def tea_transform(buf: list[int], inp: list[int]) -> None:
    DELTA: int = 0x9E3779B9
    sum_val: int = 0
    b0: int = buf[0]
    b1: int = buf[1]
    a: int = inp[0]
    b: int = inp[1]
    c: int = inp[2]
    d: int = inp[3]

    for _ in range(16):
        sum_val = (sum_val + DELTA) & 0xFFFFFFFF
        b0 = (b0 + (((b1 << 4) + a) ^ (b1 + sum_val) ^ ((b1 >> 5) + b))) & 0xFFFFFFFF
        b1 = (b1 + (((b0 << 4) + c) ^ (b0 + sum_val) ^ ((b0 >> 5) + d))) & 0xFFFFFFFF

    buf[0] = (buf[0] + b0) & 0xFFFFFFFF
    buf[1] = (buf[1] + b1) & 0xFFFFFFFF


@final
class DXRoot(DXEntriesBase):
    __slots__ = ()
    _pack_ = 1

    def __init__(self, inode: "Directory") -> None:
        super().__init__(inode, 0)

    # _anonymous_ = ("")
    _fields_ = [
        ("dot", DotDirectoryEntry2),
        ("dotdot", DotDirectoryEntry2),
        ("dx_root_info", DXRootInfo),
        ("limit", c_uint16),
        ("count", c_uint16),
        ("block", c_uint32),
        # ("entries", DXEntry * self.count),
    ]

    def compute_dx_hash(self, name: bytes, hash_version: int) -> tuple[int, int]:
        seed = self.directory.volume.superblock.s_hash_seed  # pyright: ignore[reportAny]
        buf: list[int] = [0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476]
        if seed is not None and any(seed):  # pyright: ignore[reportAny]
            buf = list(seed)  # pyright: ignore[reportAny]

        match hash_version:
            case DX_HASH.LEGACY | DX_HASH.LEGACY_UNSIGNED:
                hash0: int = 0x12A3FE2D
                hash1: int = 0x37ABE8F9
                for i in range(len(name)):
                    byte_val = sign_extended_byte(
                        hash_version == DX_HASH.LEGACY, name[i]
                    )
                    hash_val: int = hash1 + (hash0 ^ (byte_val * 7152373))
                    if hash_val & 0x80000000:
                        hash_val -= 0x7FFFFFFF

                    hash1 = hash0
                    hash0 = hash_val & 0xFFFFFFFF

                hash_val = (hash0 << 1) & 0xFFFFFFFF
                minor_hash = 0

            case DX_HASH.HALF_MD4 | DX_HASH.HALF_MD4_UNSIGNED:
                signed_hash = hash_version == DX_HASH.HALF_MD4
                p = 0
                length = len(name)
                while length > 0:
                    inp = str2hashbuf(name[p:], length, 8, signed_hash)
                    half_md4_transform(buf, inp)
                    length -= 32
                    p += 32

                hash_val = buf[1]
                minor_hash = buf[2]

            case DX_HASH.TEA | DX_HASH.TEA_UNSIGNED:
                signed_hash = hash_version == DX_HASH.TEA
                p = 0
                length = len(name)
                while length > 0:
                    inp = str2hashbuf(name[p:], length, 4, signed_hash)
                    tea_transform(buf, inp)
                    length -= 16
                    p += 16

                hash_val = buf[0]
                minor_hash = buf[1]

            case DX_HASH.SIPHASH:
                raise HtreeHashError("SipHash not yet supported for htree lookup")

            case _:
                raise HtreeHashError(f"Unknown hash version: {hash_version}")

        hash_val &= ~1
        if hash_val == 0xFFFFFFFE:
            hash_val = 0xFFFFFFFC

        return hash_val, minor_hash

    def lookup(self, name: str | bytes) -> int | None:
        if isinstance(name, str):
            name = name.encode("utf-8")

        hash_version = assert_cast(self.dx_root_info.hash_version, int)  # pyright: ignore[reportAny]
        hash_val, _minor_hash = self.compute_dx_hash(name, hash_version)
        block_num: int = assert_cast(self.block, int)  # pyright: ignore[reportAny]  # default block
        entries = self.entries
        lo: int = 0
        hi: int = len(entries) - 1
        while lo <= hi:
            mid: int = (lo + hi) // 2
            if entries[mid].hash <= hash_val:  # pyright: ignore[reportAny]
                lo = mid + 1

            else:
                hi = mid - 1

        leaf_entries: DXEntries | None = None
        leaf_idx: int = -1
        if hi >= 0:
            block_num = entries[hi].block  # pyright: ignore[reportAny]
            leaf_entries = entries
            leaf_idx = hi

        for _ in range(assert_cast(self.dx_root_info.indirect_levels, int)):  # pyright: ignore[reportAny]
            node = DXNode(
                self.directory,
                block_num * self.directory.block_size,
            )
            node_entries = node.entries

            lo = 0
            hi = len(node_entries) - 1
            while lo <= hi:
                mid = (lo + hi) // 2
                if node_entries[mid].hash <= hash_val:  # pyright: ignore[reportAny]
                    lo = mid + 1

                else:
                    hi = mid - 1

            block_num = (
                node_entries[hi].block  # pyright: ignore[reportAny]
                if hi >= 0
                else assert_cast(node.block, int)  # pyright: ignore[reportAny]
            )

            if hi >= 0:
                leaf_entries = node_entries
                leaf_idx = hi

            else:
                leaf_entries = None
                leaf_idx = -1

        blocks_to_scan: list[int] = []
        if leaf_entries is not None and leaf_idx >= 0:
            current_idx = leaf_idx
            while current_idx < len(leaf_entries):
                entry = leaf_entries[current_idx]
                blocks_to_scan.append(entry.block)  # pyright: ignore[reportAny]
                next_idx = current_idx + 1
                if (
                    next_idx >= len(leaf_entries)
                    or not (leaf_entries[next_idx].hash & 1)  # pyright: ignore[reportAny]
                    or (leaf_entries[next_idx].hash & ~1) != (hash_val & ~1)  # pyright: ignore[reportAny]
                ):
                    break

                current_idx = next_idx
        else:
            blocks_to_scan = [block_num]

        has_filetype = self.directory.has_filetype
        with BlockIO(self.directory) as blockio:
            for scan_block in blocks_to_scan:
                leaf_data = blockio.blocks[scan_block]
                offset: int = 0
                while offset + 8 <= len(leaf_data):
                    inode_val = int.from_bytes(leaf_data[offset : offset + 4], "little")
                    rec_len = int.from_bytes(
                        leaf_data[offset + 4 : offset + 6], "little"
                    )

                    if rec_len == 0:
                        break

                    name_len = (
                        leaf_data[offset + 6]
                        if has_filetype
                        else int.from_bytes(
                            leaf_data[offset + 6 : offset + 8], "little"
                        )
                    )

                    if inode_val == 0 or name_len == 0:
                        offset += rec_len
                        continue

                    name_start = offset + 8
                    if name_start + name_len > len(leaf_data):
                        break

                    entry_name = leaf_data[name_start : name_start + name_len]
                    if entry_name == name:
                        return inode_val

                    offset += rec_len

        return None


@final
class DXFake(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    # _anonymous_ = ("")
    _fields_ = [
        ("inode", c_uint32),  # 0
        ("rec_len", c_uint16),
    ]

    @property
    def expected_magic(self) -> int:
        return 0

    @property
    def magic(self) -> int:
        inode = assert_cast(self.inode, int)  # pyright: ignore[reportAny]
        return inode


@final
class DXNode(DXEntriesBase):
    __slots__ = ()

    _pack_ = 1
    # _anonymous_ = ("")
    _fields_ = [
        ("fake", DXFake),
        ("name_len", c_uint8),
        ("file_type", c_uint8),
        ("limit", c_uint16),
        ("count", c_uint16),
        ("block", c_uint32),
        # ("entries", DXEntry * self.count),
    ]

    def __init__(self, directory: "Directory", offset: int) -> None:
        super().__init__(directory, offset)


@final
class DXTail(DXBase):
    __slots__ = ("parent",)

    _pack_ = 1
    # _anonymous_ = ("dt_reserved")
    _fields_ = [
        ("dt_reserved", c_uint32),
        ("dt_checksum", c_uint16),
    ]

    def __init__(self, parent: DXNode) -> None:
        self.parent = parent
        count = assert_cast(parent.count, int)  # pyright: ignore[reportAny]
        super().__init__(
            parent.directory,
            parent.offset + parent.size + (count + 1) * parent.info_length,
        )



from .block import (
    BlockIO,
    BlockIOBlocks,
)
from .blockdescriptor import BlockDescriptor
from .directory import (
    EXT4_DIR_PAD,
    EXT4_DIR_ROUND,
    EXT4_MAX_REC_LEN,
    EXT4_NAME_LEN,
    DirectoryEntry,
    DirectoryEntry2,
    DirectoryEntryHash,
    DirectoryEntryTail,
)
from .enum import (
    DX_HASH,
    EXT2_FLAGS,
    EXT4_BG,
    EXT4_CHKSUM,
    EXT4_DEFM,
    EXT4_ERRORS,
    EXT4_FEATURE_COMPAT,
    EXT4_FEATURE_INCOMPAT,
    EXT4_FEATURE_RO_COMPAT,
    EXT4_FL,
    EXT4_FS,
    EXT4_FT,
    EXT4_INO,
    EXT4_MOUNT,
    EXT4_MOUNT2,
    EXT4_OS,
    EXT4_REV,
    FS_ENCRYPTION_MODE,
    MODE,
)
from .extent import (
    Extent,
    ExtentBlocks,
    ExtentHeader,
    ExtentIndex,
    ExtentTail,
)
from .htree import (
    DotDirectoryEntry2,
    DXEntry,
    DXRoot,
    DXRootInfo,
)
from .inode import (
    BlockDevice,
    CharacterDevice,
    Directory,
    Fifo,
    File,
    Hurd1,
    Hurd2,
    Inode,
    InodeError,
    Linux1,
    Linux2,
    Masix1,
    Masix2,
    Osd1,
    Osd2,
    Socket,
    SymbolicLink,
)
from .struct import (
    ChecksumError,
    MagicError,
)
from .superblock import Superblock
from .volume import (
    InvalidStreamException,
    Volume,
)
from .xattr import (
    ExtendedAttributeEntry,
    ExtendedAttributeError,
    ExtendedAttributeHeader,
    ExtendedAttributeIBodyHeader,
)

__all__ = [
    "BlockDescriptor",
    "BlockDevice",
    "BlockIO",
    "BlockIOBlocks",
    "CharacterDevice",
    "ChecksumError",
    "Directory",
    "DirectoryEntry",
    "DirectoryEntry2",
    "DirectoryEntryHash",
    "DirectoryEntryTail",
    "DotDirectoryEntry2",
    "DX_HASH",
    "DXEntry",
    "DXRoot",
    "DXRootInfo",
    "EXT2_FLAGS",
    "EXT4_BG",
    "EXT4_CHKSUM",
    "EXT4_DEFM",
    "EXT4_DIR_PAD",
    "EXT4_DIR_ROUND",
    "EXT4_ERRORS",
    "EXT4_FEATURE_COMPAT",
    "EXT4_FEATURE_INCOMPAT",
    "EXT4_FEATURE_RO_COMPAT",
    "EXT4_FL",
    "EXT4_FS",
    "EXT4_FT",
    "EXT4_INO",
    "EXT4_MAX_REC_LEN",
    "EXT4_MOUNT",
    "EXT4_MOUNT2",
    "EXT4_NAME_LEN",
    "EXT4_OS",
    "EXT4_REV",
    "ExtendedAttributeEntry",
    "ExtendedAttributeError",
    "ExtendedAttributeHeader",
    "ExtendedAttributeIBodyHeader",
    "Extent",
    "ExtentBlocks",
    "ExtentHeader",
    "ExtentIndex",
    "ExtentTail",
    "Fifo",
    "File",
    "FS_ENCRYPTION_MODE",
    "Hurd1",
    "Hurd2",
    "Inode",
    "InodeError",
    "InvalidStreamException",
    "Linux1",
    "Linux2",
    "MagicError",
    "Masix1",
    "Masix2",
    "MODE",
    "Osd1",
    "Osd2",
    "Socket",
    "Superblock",
    "SymbolicLink",
    "Volume",
]



from collections.abc import Iterator
from ctypes import (
    c_uint16,
    c_uint32,
    sizeof,
)
from typing import (
    TYPE_CHECKING,
    final,
)

from ._compat import assert_cast
from .struct import (
    Ext4Struct,
    crc32c,
)

if TYPE_CHECKING:
    from .inode import Inode
    from .volume import Volume


@final
class ExtentBlocks:
    __slots__: tuple[str, ...] = (
        "_null_block",
        "extent",
    )

    def __init__(self, extent: "Extent") -> None:
        self.extent: Extent = extent
        self._null_block: bytearray = bytearray(self.block_size)

    @property
    def block_size(self) -> int:
        return self.extent.block_size

    @property
    def volume(self) -> "Volume":
        return self.extent.volume

    @property
    def ee_start(self) -> int:
        return self.extent.ee_start

    @property
    def ee_block(self) -> int:
        ee_block = assert_cast(self.extent.ee_block, int)  # pyright: ignore[reportAny]
        return ee_block

    @property
    def ee_len(self) -> int:
        # Don't use ee_len as we want to know the value for
        # uninitialized blocks as well
        return self.extent.len

    @property
    def is_initialized(self) -> bool:
        return self.extent.is_initialized

    def __contains__(self, ee_block: int) -> bool:
        return self.ee_block <= ee_block < self.ee_block + self.ee_len

    def __getitem__(self, ee_block: int) -> bytearray | bytes:
        block_size = self.block_size
        if not self.is_initialized or ee_block not in self:
            # Uninitialized
            return self._null_block

        disk_block: int = self.ee_start + (ee_block - self.ee_block)
        _ = self.volume.seek(disk_block * block_size)
        return self.volume.read(block_size)

    def __iter__(self) -> Iterator[int]:
        return iter(range(self.ee_block, self.ee_block + self.ee_len))

    def __len__(self) -> int:
        return self.ee_len


@final
class ExtentHeader(Ext4Struct):
    __slots__ = (
        "extents",
        "indices",
        "tail",
        "tree",
    )

    _pack_ = 1
    # _anonymous_ = ()
    _fields_ = [
        ("eh_magic", c_uint16),
        ("eh_entries", c_uint16),
        ("eh_max", c_uint16),
        ("eh_depth", c_uint16),
        ("eh_generation", c_uint32),
    ]

    def __init__(self, tree: "ExtentTree", offset: int) -> None:
        self.tree: ExtentTree = tree
        super().__init__(self.inode.volume, offset)

        self.indices: list[ExtentIndex] = []
        self.extents: list[Extent] = []

        offset = self.offset + self.size
        eh_entries = assert_cast(self.eh_entries, int)  # pyright: ignore[reportAny]
        for i in range(0, eh_entries):
            if self.eh_depth == 0:
                self.extents.append(Extent(self, offset, i))
                offset += sizeof(Extent)

            else:
                self.indices.append(ExtentIndex(self, offset, i))
                offset += sizeof(ExtentIndex)

        self.indices.sort(key=lambda entry: entry.ei_no)
        self.extents.sort(key=lambda entry: entry.ee_no)
        i_block = type(self.inode).i_block
        i_block_offset = self.inode.offset + i_block.offset
        self.tail = (
            ExtentTail(self, offset)
            if offset in range(i_block_offset, i_block_offset + i_block.size)
            else None
        )

    @property
    def inode(self) -> "Inode":
        return self.tree.inode

    @Ext4Struct.expected_magic.getter
    def expected_magic(self) -> int:
        return 0xF30A

    @Ext4Struct.magic.getter
    def magic(self) -> int:
        eh_magic = assert_cast(self.eh_magic, int)  # pyright: ignore[reportAny]
        return eh_magic

    @Ext4Struct.expected_checksum.getter
    def expected_checksum(self) -> int | None:
        if self.tail is None:
            return None

        et_checksum = assert_cast(self.tail.et_checksum, int)  # pyright: ignore[reportAny]
        if not et_checksum:
            return None

        return et_checksum

    @property
    def seed(self) -> int:
        return self.inode.seed

    @Ext4Struct.checksum.getter
    def checksum(self) -> int | None:
        if self.expected_checksum is None:
            return None

        assert self.tail is not None
        _ = self.volume.seek(self.offset)
        data = self.volume.read(self.tail.offset - self.offset)
        return crc32c(data, self.seed)


@final
class ExtentIndex(Ext4Struct):
    __slots__ = (
        "ei_no",
        "header",
    )

    _pack_ = 1
    # _anonymous_ = ("ei_unused",)
    _fields_ = [
        ("ei_block", c_uint32),
        ("ei_leaf_lo", c_uint32),
        ("ei_leaf_hi", c_uint16),
        ("ei_unused", c_uint16),
    ]

    def __init__(self, header: ExtentHeader, offset: int, ei_no: int) -> None:
        self.ei_no: int = ei_no
        self.header: ExtentHeader = header
        super().__init__(self.inode.volume, offset)

    @property
    def ei_leaf(self) -> int:
        ei_leaf_lo = assert_cast(self.ei_leaf_lo, int)  # pyright: ignore[reportAny]
        ei_leaf_hi = assert_cast(self.ei_leaf_hi, int)  # pyright: ignore[reportAny]
        return ei_leaf_hi << 32 | ei_leaf_lo

    @property
    def tree(self) -> "ExtentTree":
        return self.header.tree

    @property
    def inode(self) -> "Inode":
        return self.tree.inode


@final
class Extent(Ext4Struct):
    __slots__ = (
        "blocks",
        "ee_no",
        "header",
    )

    _pack_ = 1
    # _anonymous_ = ("ei_unused",)
    _fields_ = [
        ("ee_block", c_uint32),
        ("ee_len", c_uint16),
        ("ee_start_hi", c_uint16),
        ("ee_start_lo", c_uint32),
    ]

    def __init__(self, header: ExtentHeader, offset: int, ee_no: int) -> None:
        super().__init__(header.inode.volume, offset)
        self.ee_no: int = ee_no
        self.header: ExtentHeader = header
        self.blocks: ExtentBlocks = ExtentBlocks(self)

    @property
    def ee_start(self) -> int:
        ee_start_lo = assert_cast(self.ee_start_lo, int)  # pyright: ignore[reportAny]
        ee_start_hi = assert_cast(self.ee_start_hi, int)  # pyright: ignore[reportAny]
        return ee_start_hi << 32 | ee_start_lo

    @property
    def tree(self) -> "ExtentTree":
        return self.header.tree

    @property
    def is_initialized(self) -> bool:
        return self.ee_len < 32768  # pyright: ignore[reportAny]

    @property
    def len(self) -> int:
        return self.ee_len if self.is_initialized else self.ee_len - 32768  # pyright: ignore[reportAny]

    @property
    def inode(self) -> "Inode":
        return self.tree.inode

    @property
    def block_size(self) -> int:
        return self.volume.block_size

    def read(self) -> bytes:
        return b"".join(self.blocks[b] for b in self.blocks)


@final
class ExtentTail(Ext4Struct):
    __slots__ = ("header",)

    _pack_ = 1
    _fields_ = [
        ("et_checksum", c_uint32),
    ]

    def __init__(self, header: ExtentHeader, offset: int) -> None:
        self.header: ExtentHeader = header
        super().__init__(self.inode.volume, offset)

    @property
    def tree(self) -> "ExtentTree":
        return self.header.tree

    @property
    def inode(self) -> "Inode":
        return self.tree.inode


class ExtentTree:
    __slots__: tuple[str, ...] = (
        "headers",
        "inode",
    )

    def __init__(self, inode: "Inode") -> None:
        self.inode: Inode = inode
        self.headers: list[ExtentHeader] = []
        if not self.has_extents:
            return

        to_process = [self.offset]
        while to_process:
            header_offset = to_process.pop(0)
            header = ExtentHeader(self, header_offset)
            self.headers.append(header)
            for index in header.indices:
                to_process.append(index.ei_leaf * self.volume.block_size)

    @property
    def volume(self) -> "Volume":
        return self.inode.volume

    @property
    def offset(self) -> int:
        return self.inode.offset + type(self.inode).i_block.offset

    @property
    def has_extents(self) -> bool:
        return self.inode.has_extents

    def verify(self) -> None:
        pass

    def validate(self) -> None:
        for header in self.headers:
            header.validate()

    @property
    def extents(self) -> list[Extent]:
        extents: list[Extent] = []
        for header in self.headers:
            extents += header.extents

        return extents

    @property
    def indices(self) -> list[ExtentIndex]:
        indices: list[ExtentIndex] = []
        for header in self.headers:
            indices += header.indices

        return indices
